"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import json
from collections import deque
from copy import deepcopy
from typing import Any, Iterator

import regex as re

from cfnlint.context.conditions.exceptions import Unsatisfiable
from cfnlint.helpers import (
    AVAILABILITY_ZONES,
    PSEUDOPARAMS,
    REGEX_SUB_PARAMETERS,
    REGIONS,
    is_function,
)
from cfnlint.jsonschema import ValidationError, Validator
from cfnlint.jsonschema._typing import ResolutionResult
from cfnlint.jsonschema._utils import equal


def unresolvable(validator: Validator, instance: Any) -> ResolutionResult:
    return
    yield


def ref(validator: Validator, instance: Any) -> ResolutionResult:
    if not isinstance(instance, (str, dict)):
        return

    for instance, instance_validator, _ in validator.resolve_value(instance):
        if validator.is_type(instance, "string"):
            # if the ref is to pseudo-parameter or parameter we can validate the values
            for v, c in instance_validator.context.ref_value(instance):
                yield v, instance_validator.evolve(context=c), None
            return


def find_in_map(validator: Validator, instance: Any) -> ResolutionResult:
    if not validator.is_type(instance, "array"):
        return
    if len(instance) not in [3, 4]:
        return

    default_value_found = False
    if len(instance) == 4:
        options = instance[3]
        if validator.is_type(options, "object"):
            if "DefaultValue" in options:
                default_value_found = True
                for value, v, _ in validator.resolve_value(options["DefaultValue"]):
                    yield value, v.evolve(
                        context=v.context.evolve(
                            path=v.context.path.evolve(
                                value_path=deque([4, "DefaultValue"])
                            )
                        ),
                    ), None

    if not default_value_found and not validator.context.mappings.maps:
        if validator.context.mappings.is_transform:
            return
        yield None, validator, ValidationError(
            (
                f"{instance[0]!r} is not one of "
                f"{list(validator.context.mappings.maps.keys())!r}"
            ),
            path=deque([0]),
        )

    mappings = list(validator.context.mappings.maps.keys())
    results = []
    found_valid_combination = False
    k, v = is_function(instance[0])
    if k == "Ref" and v in PSEUDOPARAMS:
        return
    for map_name, map_v, _ in validator.resolve_value(instance[0]):
        if not validator.is_type(map_name, "string"):
            continue

        if all(not (equal(map_name, each)) for each in mappings):
            if not default_value_found:
                # Validated again as Fn::Transform can be used along side
                # other key/values
                if validator.context.mappings.is_transform:
                    continue
                results.append(
                    (
                        None,
                        map_v,
                        ValidationError(
                            f"{map_name!r} is not one of {mappings!r}",
                            path=deque([0]),
                        ),
                    )
                )
            continue

        if validator.context.mappings.maps[map_name].is_transform:
            continue

        k, v = is_function(instance[2])
        if k == "Ref" and v in PSEUDOPARAMS:
            continue

        k, v = is_function(instance[1])
        if k == "Ref" and v in PSEUDOPARAMS:
            if isinstance(instance[2], str):
                found_top_level_key = False
                found_second_key = False
                for top_level_key, top_values in validator.context.mappings.maps[
                    map_name
                ].keys.items():
                    if v == "AWS::AccountId":
                        if not re.match("^[0-9]{12}$", top_level_key):
                            continue
                    elif v == "AWS::Region":
                        if top_level_key not in REGIONS:
                            continue
                    found_top_level_key = True
                    for second_level_key, second_v, _ in validator.resolve_value(
                        instance[2]
                    ):
                        if second_level_key in top_values.keys:
                            for value in validator.context.mappings.maps[
                                map_name
                            ].find_in_map(
                                top_level_key,
                                second_level_key,
                            ):
                                found_second_key = True
                                yield (
                                    value,
                                    validator.evolve(
                                        context=validator.context.evolve(
                                            path=validator.context.path.evolve(
                                                value_path=deque(
                                                    [
                                                        "Mappings",
                                                        map_name,
                                                        top_level_key,
                                                        second_level_key,
                                                    ]
                                                )
                                            )
                                        )
                                    ),
                                    None,
                                )

                if not found_top_level_key:
                    yield None, validator, ValidationError(
                        (
                            f"{instance[1]!r} is not a "
                            f"first level key for mapping {map_name!r}"
                        ),
                        path=deque([1]),
                    )
                elif not found_second_key:
                    yield None, validator, ValidationError(
                        (
                            f"{instance[2]!r} is not a "
                            "second level key when "
                            f"{instance[1]!r} is resolved "
                            f"for mapping {map_name!r}"
                        ),
                        path=deque([2]),
                    )
            continue

        for top_level_key, top_v, _ in validator.resolve_value(instance[1]):
            if validator.is_type(top_level_key, "integer"):
                top_level_key = str(top_level_key)
            if not validator.is_type(top_level_key, "string"):
                continue

            top_level_keys = list(validator.context.mappings.maps[map_name].keys.keys())
            if all(not (equal(top_level_key, each)) for each in top_level_keys):
                if not default_value_found:
                    results.append(
                        (
                            None,
                            top_v,
                            ValidationError(
                                (
                                    f"{top_level_key!r} is not one of "
                                    f"{top_level_keys!r} for mapping "
                                    f"{map_name!r}"
                                ),
                                path=deque([1]),
                            ),
                        )
                    )
                continue

            if (
                not top_level_key
                or validator.context.mappings.maps[map_name]
                .keys[top_level_key]
                .is_transform
            ):
                continue

            top_v = validator.evolve(
                context=top_v.context.evolve(resolve_pseudo_parameters=False)
            )
            for second_level_key, second_v, err in top_v.resolve_value(instance[2]):
                if validator.is_type(second_level_key, "integer"):
                    second_level_key = str(second_level_key)
                if not validator.is_type(second_level_key, "string"):
                    continue
                second_level_keys = list(
                    validator.context.mappings.maps[map_name]
                    .keys[top_level_key]
                    .keys.keys()
                )
                if all(
                    not (equal(second_level_key, each)) for each in second_level_keys
                ):
                    if not default_value_found:
                        results.append(
                            (
                                None,
                                second_v,
                                ValidationError(
                                    (
                                        f"{second_level_key!r} is not "
                                        f"one of {second_level_keys!r} "
                                        f"for mapping {map_name!r} and "
                                        f"key {top_level_key!r}"
                                    ),
                                    path=deque([2]),
                                ),
                            )
                        )
                    continue

                found_valid_combination = True

                for value in validator.context.mappings.maps[map_name].find_in_map(
                    top_level_key,
                    second_level_key,
                ):
                    yield (
                        value,
                        second_v.evolve(
                            context=second_v.context.evolve(
                                path=second_v.context.path.evolve(
                                    value_path=deque(
                                        [
                                            "Mappings",
                                            map_name,
                                            top_level_key,
                                            second_level_key,
                                        ]
                                    )
                                )
                            )
                        ),
                        None,
                    )

    if not found_valid_combination:
        yield from iter(results)


def get_azs(validator: Validator, instance: Any) -> ResolutionResult:
    if not isinstance(instance, (str, dict)):
        return

    for instance, v, _ in validator.resolve_value(instance):
        if v.is_type(instance, "string"):
            if instance == "":
                for region in v.context.regions:
                    yield (
                        AVAILABILITY_ZONES.get(region),
                        v,
                        None,
                    )
            # if the ref is to pseudo-parameter or parameter we can validate the values
            elif instance in AVAILABILITY_ZONES:
                yield AVAILABILITY_ZONES.get(instance), v, None


def _join_expansion(
    validator: Validator, instances: Any
) -> Iterator[tuple[Any, Validator]]:
    if len(instances) == 0:
        return

    if len(instances) == 1:
        for value, instance_validator, _ in validator.resolve_value(instances[0]):
            if not isinstance(value, (str, int, float, bool)):
                raise ValueError(f"Incorrect value type for {value!r}")
            yield [value], instance_validator
        return

    for value, instance_validator, errs in validator.resolve_value(instances[0]):
        if errs:
            continue
        if not isinstance(value, (str, int, float, bool)):
            raise ValueError(f"Incorrect value type for {value!r}")
        for values, expansion_validator in _join_expansion(
            instance_validator, instances[1:]
        ):
            yield [value] + values, expansion_validator


def join(validator: Validator, instance: Any) -> ResolutionResult:
    # quick validations
    if not validator.is_type(instance, "array"):
        return
    if not len(instance) == 2:
        return

    for delimiter, delimiter_v, _ in validator.resolve_value(instance[0]):
        if not delimiter_v.is_type(delimiter, "string"):
            continue
        for values, values_v, _ in delimiter_v.resolve_value(instance[1]):
            if not values_v.is_type(values, "array"):
                continue
            try:
                for value, expansion_validator in _join_expansion(values_v, values):
                    # reset the value path because we could get a Ref value_path
                    yield delimiter.join(
                        [str(v) for v in value]
                    ), expansion_validator.evolve(
                        context=expansion_validator.context.evolve(
                            path=expansion_validator.context.path.evolve(
                                value_path=deque([])
                            )
                        )
                    ), None
            except (ValueError, TypeError):
                return


def select(validator: Validator, instance: Any) -> ResolutionResult:
    # quick validations
    if not validator.is_type(instance, "array"):
        return
    if not len(instance) == 2:
        return

    # get the values from the list
    indexes = validator.resolve_value(instance[0])
    objs = validator.resolve_value(instance[1])

    for i, _, _ in indexes:
        for obj, obj_v, _ in objs:
            try:
                i = int(i)
            except ValueError:
                continue
            if not validator.is_type(obj, "array"):
                continue
            if len(obj) <= i:
                continue
            yield from obj_v.resolve_value(obj[i])


def split(validator: Validator, instance: Any) -> ResolutionResult:
    if not validator.is_type(instance, "array"):
        return
    if not len(instance) == 2:
        return

    for delimiter, _, _ in validator.resolve_value(instance[0]):
        for source_string, source_v, _ in validator.resolve_value(instance[1]):
            if not source_v.is_type(delimiter, "string"):
                continue
            if not source_v.is_type(source_string, "string"):
                continue

            yield source_string.split(delimiter), source_v, None


def _sub_parameter_expansion(
    validator: Validator, parameters: dict[str, Any]
) -> Iterator[dict[str, Any]]:
    parameters = parameters.copy()
    if len(parameters) == 0:
        yield {}
        return

    if len(parameters) == 1:
        for key, value in parameters.items():
            for resolved_value, _, _ in validator.resolve_value(value):
                yield {key: resolved_value}
        return

    key = list(parameters.keys())[0]
    value = parameters.pop(key)
    for resolved_value, _, _ in validator.resolve_value(value):
        for values in _sub_parameter_expansion(validator, parameters):
            yield dict({key: resolved_value}, **values)


def _sub_string(validator: Validator, string: str) -> ResolutionResult:
    sub_regex = re.compile(r"(\${([^!].*?)})")

    def _replace(matchobj):
        nonlocal validator
        for value, c in validator.context.ref_value(matchobj.group(2).strip()):
            if not isinstance(value, (str, int, float, bool)):
                raise ValueError(f"Parameter {matchobj.group(2)!r} has wrong type")

            validator = validator.evolve(
                context=validator.context.evolve(
                    ref_values=c.ref_values,
                )
            )
            return str(value)
        raise ValueError(f"No matches for {matchobj.group(2)!r}")

    try:
        yield re.sub(sub_regex, _replace, string), validator, None
    except ValueError:
        return


def sub(validator: Validator, instance: Any) -> ResolutionResult:
    if not (
        validator.is_type(instance, "array") or validator.is_type(instance, "string")
    ):
        return

    if validator.is_type(instance, "array"):
        if len(instance) != 2:
            return

        string = instance[0]
        parameters = deepcopy(instance[1])
        if not validator.is_type(string, "string"):
            return
        if not validator.is_type(parameters, "object"):
            return

        sub_parameters = REGEX_SUB_PARAMETERS.findall(string)
        for parameter in sub_parameters:
            if parameter in parameters:
                continue
            if "." in parameter:
                parameters[parameter] = {"Fn::GetAtt": parameter}
            else:
                parameters[parameter] = {"Ref": parameter}

        for resolved_parameters in _sub_parameter_expansion(validator, parameters):
            resolved_validator = validator.evolve(
                context=validator.context.evolve(
                    ref_values=resolved_parameters,
                )
            )
            for sub_string, sub_validator, _ in _sub_string(resolved_validator, string):
                for key in instance[1].keys():
                    sub_validator.context.ref_values.pop(key, None)

                yield sub_string, sub_validator, None

        return

    # its a string
    sub_parameters = REGEX_SUB_PARAMETERS.findall(instance)
    parameters = {}
    for parameter in sub_parameters:
        if "." in parameter:
            parameters[parameter] = {"Fn::GetAtt": parameter}
        else:
            parameters[parameter] = {"Ref": parameter}
    for resolved_parameters in _sub_parameter_expansion(validator, parameters):
        resolved_validator = validator.evolve(
            context=validator.context.evolve(
                ref_values=resolved_parameters,
            )
        )
        yield from _sub_string(resolved_validator, instance)


def if_(validator: Validator, instance: Any) -> ResolutionResult:
    if not validator.is_type(instance, "array"):
        return

    if len(instance) != 3:
        return

    for i in [1, 2]:
        for value, v, err in validator.resolve_value(instance[i]):
            try:
                yield (
                    value,
                    v.evolve(
                        context=v.context.evolve(
                            path=v.context.path.evolve(value_path=deque([i])),
                            conditions=v.context.conditions.evolve(
                                {instance[0]: True if i == 1 else False}
                            ),
                        ),
                    ),
                    err,
                )
            except Unsatisfiable:
                continue


def to_json_string(validator: Validator, instance: Any) -> ResolutionResult:
    for value, v, err in validator.resolve_value(instance):
        yield json.dumps(value), v, err


# not all functions need to be resolved.  These functions
# allow us to pull up values from nested functions
# allowing us to test the possible values against the schema
fn_resolvers: dict[str, Any] = {
    "Fn::Base64": unresolvable,
    "Fn::Cidr": unresolvable,
    "Fn::FindInMap": find_in_map,
    "Fn::ForEach": unresolvable,
    "Fn::GetAtt": unresolvable,
    "Fn::GetAZs": get_azs,
    "Fn::ImportValue": unresolvable,
    "Fn::If": if_,
    "Fn::Join": join,
    "Fn::Select": select,
    "Fn::Split": split,
    "Fn::Sub": sub,
    "Fn::Transform": unresolvable,
    "Fn::ToJsonString": to_json_string,
    "Fn::Equals": unresolvable,
    "Fn::Or": unresolvable,
    "Fn::And": unresolvable,
    "Fn::Not": unresolvable,
    "Condition": unresolvable,
    "Fn::Length": unresolvable,
    "Ref": ref,
}



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import MagicMock, call, mock_open, patch

from cfnlint.schema._patch import SchemaPatch
from cfnlint.schema.manager import ProviderSchemaManager, ResourceNotFoundError

LOGGER = logging.getLogger("cfnlint.schema.manager")
LOGGER.disabled = True


class TestUpdateResourceSchemas(BaseTestCase):
    """Used for Testing Resource Schemas"""

    def setUp(self) -> None:
        super().setUp()

        self.schema_zip = "test/fixtures/registry/schema.zip"
        self.manager = ProviderSchemaManager()
        self.schemas = dict.fromkeys(["aws-lambda-codesigningconfig"])
        for resource in self.schemas:
            with open(f"test/fixtures/registry/schemas/{resource}.json") as fh:
                self.schemas[resource] = fh.read()

        self.schema_patch = [{"op": "add", "path": "/cfnSchema", "value": ["test"]}]

    @patch("cfnlint.schema.manager.url_has_newer_version")
    @patch("cfnlint.schema.manager.json.dump")
    @patch("cfnlint.schema.manager.REGIONS", ["us-east-1"])
    @patch("cfnlint.schema.manager.get_url_retrieve")
    @patch("cfnlint.schema.manager.zipfile.ZipFile")
    @patch("cfnlint.schema.manager.os.listdir")
    @patch("cfnlint.schema.manager.os.path.isfile")
    @patch("cfnlint.schema.manager.os.remove")
    @patch("cfnlint.schema.manager.os.walk")
    @patch("cfnlint.schema.manager.filecmp.cmp")
    @patch("cfnlint.schema.manager.load_resource")
    @patch("cfnlint.schema.manager.shutil.rmtree")
    def test_update_resource_spec(
        self,
        mock_shutil_rmtree,
        mock_load_resource,
        mock_filecmp_cmp,
        mock_os_walk,
        mock_os_remove,
        mock_os_path_isfile,
        mock_os_listdir,
        mock_zipfile,
        mock_get_url_retrieve,
        mock_json_dump,
        mock_url_newer_version,
    ):
        schema = self.schemas["aws-lambda-codesigningconfig"]
        schema_json = json.loads(schema)

        """Success update resource spec"""
        mock_url_newer_version.return_value = True
        mock_get_url_retrieve.return_value = self.schema_zip
        mock_zipfile.return_value = MagicMock()
        mock_os_listdir.return_value = [
            "aws_lambda_codesigningconfig.json",
            "__init__.py",
        ]
        mock_os_path_isfile.side_effect = [True, True, True, True]
        mock_load_resource.return_value = self.schema_patch
        mock_os_walk.return_value = iter(
            [("all", [], ["aws_lambda_codesigningconfig.json"])]
        )
        builtin_module_name = "builtins"

        with patch(
            "{}.open".format(builtin_module_name), mock_open(read_data=schema)
        ) as mock_builtin_open:
            self.manager._update_provider_schema("us-east-1", False)
            schema_patched = {**schema_json, **{"cfnSchema": ["test"]}}
            mock_json_dump.assert_called_with(
                schema_patched,
                mock_builtin_open.return_value.__enter__.return_value,
                indent=1,
                separators=(",", ": "),
                sort_keys=True,
            )
            mock_os_listdir.assert_has_calls(
                [
                    call(f"{self.manager._root.path_relative}/us_east_1/"),
                ]
            )
            mock_zipfile.assert_has_calls([call(self.schema_zip, "r")])
            mock_filecmp_cmp.assert_not_called()
            mock_os_remove.assert_not_called()

    @patch("cfnlint.schema.manager.url_has_newer_version")
    @patch("cfnlint.schema.manager.get_url_retrieve")
    @patch("cfnlint.schema.manager.json.dump")
    @patch("cfnlint.schema.manager.REGIONS", ["us-east-1"])
    @patch("cfnlint.schema.manager.load_resource")
    @patch("cfnlint.schema.manager.zipfile.ZipFile")
    @patch("cfnlint.schema.manager.os.listdir")
    @patch("cfnlint.schema.manager.os.path.isfile")
    @patch("cfnlint.schema.manager.os.remove")
    @patch("cfnlint.schema.manager.os.walk")
    @patch("cfnlint.schema.manager.filecmp.cmp")
    @patch("cfnlint.schema.manager.shutil.rmtree")
    def test_update_resource_spec_cache(
        self,
        mock_shutil_rmtree,
        mock_filecmp_cmp,
        mock_os_walk,
        mock_os_remove,
        mock_os_path_isfile,
        mock_os_listdir,
        mock_zipfile,
        mock_load_resource,
        mock_json_dump,
        mock_get_url_retrieve,
        mock_url_newer_version,
    ):
        """Success update resource spec with cache"""
        schema = self.schemas["aws-lambda-codesigningconfig"]
        schema_json = json.loads(schema)

        mock_url_newer_version.return_value = True
        mock_get_url_retrieve.return_value = self.schema_zip
        mock_zipfile.return_value = MagicMock()
        mock_os_listdir.return_value = [
            "aws-lambda-codesigningconfig.json",
            "__init__.py",
        ]
        mock_os_path_isfile.side_effect = [True, True, True, True]
        mock_filecmp_cmp.side_effect = [True]
        mock_load_resource.return_value = self.schema_patch
        mock_os_walk.return_value = iter(
            [("all", [], ["aws_lambda_codesigningconfig.json"])]
        )
        builtin_module_name = "builtins"

        with patch(
            "{}.open".format(builtin_module_name), mock_open(read_data=schema)
        ) as mock_builtin_open:
            self.manager._update_provider_schema("us-west-2", False)
            schema_patched = {**schema_json, **{"cfnSchema": ["test"]}}
            mock_json_dump.assert_called_with(
                schema_patched,
                mock_builtin_open.return_value.__enter__.return_value,
                indent=1,
                separators=(",", ": "),
                sort_keys=True,
            )
            mock_os_listdir.assert_has_calls(
                [
                    call(f"{self.manager._root.path_relative}/us_west_2/"),
                    call(f"{self.manager._root.path_relative}/us_west_2/"),
                ]
            )
            mock_zipfile.assert_has_calls([call(self.schema_zip, "r")])
            mock_filecmp_cmp.assert_called_once()
            mock_os_remove.assert_called_once()

    @patch("cfnlint.schema.manager.json.dump")
    @patch("cfnlint.schema.manager.REGIONS", ["us-east-1"])
    @patch("cfnlint.schema.manager.os.listdir")
    @patch("cfnlint.schema.manager.os.path.isfile")
    def test_update_resource_for_iso_region(
        self,
        mock_os_path_isfile,
        mock_os_listdir,
        mock_json_dump,
    ):
        """Success update resource spec with cache"""
        schema = self.schemas["aws-lambda-codesigningconfig"]

        mock_os_listdir.return_value = [
            "aws_lambda_codesigningconfig.json",
            "__init__.py",
        ]
        mock_os_path_isfile.side_effect = [True, True, True, True]
        builtin_module_name = "builtins"

        with patch(
            "{}.open".format(builtin_module_name), mock_open(read_data=schema)
        ) as mock_builtin_open:
            self.manager._update_provider_schema("us-iso-west-1", False)
            mock_json_dump.assert_not_called()
            mock_os_listdir.assert_has_calls(
                [
                    call(f"{self.manager._root.path_relative}/us_east_1/"),
                ]
            )
            mock_builtin_open.assert_has_calls(
                [
                    call().__enter__(),
                    call().write("from __future__ import annotations\n\n"),
                    call().write(
                        "# pylint: disable=too-many-lines\ntypes: list[str] = [\n"
                    ),
                    call().write('    "AWS::CDK::Metadata",\n'),
                    call().write('    "AWS::Lambda::CodeSigningConfig",\n'),
                    call().write('    "Module",\n'),
                    call().write(
                        "]\n\n# pylint: disable=too-many-lines\ncached: list[str] = [\n"
                    ),
                    call().write('    "Module",\n'),
                    call().write('    "aws_lambda_codesigningconfig.json",\n'),
                    call().write("]\n"),
                ]
            )

    @patch("cfnlint.schema.manager.url_has_newer_version")
    @patch("cfnlint.schema.manager.get_url_retrieve")
    @patch("cfnlint.schema.manager.json.dump")
    @patch("cfnlint.schema.manager.ProviderSchemaManager._patch_provider_schema")
    @patch("cfnlint.schema.manager.REGIONS", {"us-east-1": []})
    def test_do_not_update_resource_spec(
        self,
        mock_provider_schema,
        mock_json_dump,
        mock_get_url_retrieve,
        mock_url_newer_version,
    ):
        """Success update resource spec"""

        mock_url_newer_version.return_value = False

        result = self.manager._update_provider_schema("us-east-1", False)
        self.assertIsNone(result)
        mock_get_url_retrieve.assert_not_called()
        mock_provider_schema.assert_not_called()
        mock_json_dump.assert_not_called()

    @patch("cfnlint.schema.manager.multiprocessing.Pool")
    @patch("cfnlint.schema.manager.ProviderSchemaManager._update_provider_schema")
    @patch("cfnlint.schema.manager.REGIONS", {"us-east-1": []})
    def test_update_resource_specs_python(self, mock_update_resource_spec, mock_pool):
        fake_pool = MagicMock()
        mock_pool.return_value.__enter__.return_value = fake_pool

        self.manager.update(True)

        fake_pool.starmap.assert_called_once()


class TestManagerGetResourceSchema(BaseTestCase):
    """Test get resource schema"""

    def setUp(self) -> None:
        super().setUp()

        self.manager = ProviderSchemaManager()

    def test_getting_cached_schema(self):
        rt = "AWS::EC2::VPC"

        self.manager.get_resource_schema("us-east-1", rt)
        schema = self.manager.get_resource_schema("us-east-2", rt)

        self.assertTrue(schema.is_cached)

    def test_removed_types(self):
        rt = "AWS::EC2::VPC"
        region = "us-east-1"
        self.manager.patch(SchemaPatch([], [rt], {}), region)

        with self.assertRaises(ResourceNotFoundError):
            self.manager.get_resource_schema(region, rt)

    def test_getting_us_east_1_schema_in_iso(self):
        rt = "AWS::EC2::VPC"

        schema_us_east_1 = self.manager.get_resource_schema("us-east-1", rt)
        schema_iso = self.manager.get_resource_schema("us-iso-east-1", rt)

        self.assertDictEqual(schema_us_east_1.schema, schema_iso.schema)
        self.assertTrue(schema_iso.is_cached)

    def test_type_normalization(self):

        rt = "MyCompany::MODULE"
        schema = self.manager.get_resource_schema("us-east-1", rt)

        assert schema.schema.get("typeName") == "Module"

        self.manager.get_resource_schema.cache_clear()
        self.manager._registry_schemas[rt] = True
        schema = self.manager.get_resource_schema("us-east-1", rt)
        assert schema is True


class TestManagerPatch(BaseTestCase):
    """Test patching schemas"""

    def setUp(self) -> None:
        super().setUp()
        self.manager = ProviderSchemaManager()
        self.schema_patch = [{"op": "add", "path": "/cfnSchema", "value": ["test"]}]

    @patch("cfnlint.schema.manager.print")
    @patch("cfnlint.schema.manager.sys.exit")
    def test_patch_failure(self, mock_exit, mock_print):
        """Test when patching a schema fails"""
        # Create a mock schema that will raise an exception when patch is called
        mock_schema = MagicMock()
        mock_schema.patch.side_effect = Exception("Invalid patch operation")

        # Mock get_resource_schema to return our mocked schema
        self.manager.get_resource_schema = MagicMock(return_value=mock_schema)

        # Create a patch
        resource_type = "AWS::EC2::Instance"
        patch = SchemaPatch([], [], {resource_type: self.schema_patch})

        # Apply the patch
        self.manager.patch(patch, "us-east-1")

        # Verify that print was called with the error message
        mock_print.assert_called_with(
            (
                f"Error applying patch {self.schema_patch} for "
                f"{resource_type}: Invalid patch operation"
            )
        )

        # Verify that sys.exit was called with exit code 1
        mock_exit.assert_called_with(1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.rules.formats.SubnetId import SubnetId


@pytest.fixture(scope="module")
def rule():
    rule = SubnetId()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Valid subnet id",
            "subnet-abcd1234",
            True,
        ),
        (
            "Valid subnet id long",
            "subnet-abcdefa1234567890",
            True,
        ),
        (
            "Valid but wrong type",
            [],
            True,
        ),
        ("Invalid subnet ID", "subnet-abc", False),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    result = rule.format(validator, instance)
    assert result == expected, f"Test {name!r} got {result!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.conditions.Not import Not


@pytest.fixture
def rule():
    rule = Not()
    yield rule


@pytest.fixture
def template():
    return {
        "Parameters": {
            "Environment": {
                "Type": "String",
            }
        },
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]},
            "IsProduction": {"Fn::Equals": [{"Ref": "Environment"}, "Production"]},
        },
    }


@pytest.mark.parametrize(
    "name,instance,path,errors",
    [
        (
            "Valid or with other conditions",
            {"Fn::Not": [{"Condition": "IsUsEast1"}]},
            {},
            [],
        ),
        (
            "Valid or with boolean types",
            {"Fn::Not": [True]},
            {},
            [],
        ),
        (
            "Invalid Type",
            {"Fn::Not": {}},
            {},
            [
                ValidationError(
                    "{} is not of type 'array'",
                    validator="fn_not",
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    path=deque(["Fn::Not"]),
                ),
            ],
        ),
        (
            "Integer type",
            {"Fn::Not": ["a"]},
            {},
            [
                ValidationError(
                    "'a' is not of type 'boolean'",
                    validator="fn_not",
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "items",
                            "else",
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    path=deque(["Fn::Not", 0]),
                )
            ],
        ),
        (
            "Invalid functions in Conditions",
            {"Fn::Not": [{"Fn::Contains": []}]},
            {"path": deque(["Conditions", "Condition1"])},
            [
                ValidationError(
                    "{'Fn::Contains': []} is not of type 'boolean'",
                    validator="fn_not",
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "items",
                            "else",
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    path=deque(["Fn::Not", 0]),
                )
            ],
        ),
        (
            "Valid functions in Rules",
            {"Fn::Not": [{"Fn::Contains": []}]},
            {"path": deque(["Rules", "Rule1"])},
            [],
        ),
    ],
    indirect=["path"],
)
def test_condition(name, instance, errors, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))
    assert errs == errors, name



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import jsonpatch
import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.resources.ecs.ServiceFargate import ServiceFargate


@pytest.fixture
def rule():
    rule = ServiceFargate()
    yield rule


_task_definition = {
    "Type": "AWS::ECS::TaskDefinition",
    "Properties": {
        "RequiresCompatibilities": ["FARGATE"],
    },
}

_service = {
    "Type": "AWS::ECS::Service",
    "Properties": {
        "TaskDefinition": {"Ref": "TaskDefinition"},
        "LaunchType": "FARGATE",
    },
}


@pytest.mark.parametrize(
    "template,start_path,expected",
    [
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Fn::Sub": "${TaskDefinition.Arn}"},
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Fn::GetAtt": "TaskDefinition.Arn"},
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    ("'RequiresCompatibilities' is a required property"),
                    validator="required",
                    rule=ServiceFargate(),
                    path_override=deque(["Resources", "TaskDefinition", "Properties"]),
                )
            ],
        ),
        (
            {
                "Parameters": {"MyTask": {"Type": "String"}},
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": {"Ref": "MyTask"},
                            },
                        ],
                    ),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/TaskDefinition",
                                "value": "MyTask",
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    ("'RequiresCompatibilities' is a required property"),
                    validator="required",
                    rule=ServiceFargate(),
                    path_override=deque(["Resources", "TaskDefinition", "Properties"]),
                )
            ],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": jsonpatch.apply_patch(
                        dict(_service),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/LaunchType",
                                "value": "EC2",
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/RequiresCompatibilities",
                                "value": [
                                    "EC2",
                                    "FARGATE",
                                ],
                            },
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/RequiresCompatibilities",
                                "value": [
                                    "EC2",
                                    "EXTERNAL",
                                ],
                            },
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    ("['EC2', 'EXTERNAL'] does not contain items matching 'FARGATE'"),
                    validator="contains",
                    rule=ServiceFargate(),
                    path_override=deque(
                        [
                            "Resources",
                            "TaskDefinition",
                            "Properties",
                            "RequiresCompatibilities",
                        ]
                    ),
                )
            ],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/RequiresCompatibilities",
                                "value": {"Fn::FindInMap": ["A", "B", "C"]},
                            },
                        ],
                    ),
                    "Service": dict(_service),
                }
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {"MyLaunchType": {"Type": "String"}},
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "replace",
                                "path": "/Properties/RequiresCompatibilities",
                                "value": ["EC2", {"Ref": "MyLaunchType"}],
                            },
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": "/Properties/NetworkMode",
                                "value": "awsvpc",
                            },
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {"MyNetworkMode": {"Type": "String"}},
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": "/Properties/NetworkMode",
                                "value": {"Ref": "MyNetworkMode"},
                            },
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": "/Properties/NetworkMode",
                                "value": "host",
                            },
                            {
                                "op": "remove",
                                "path": "/Properties/RequiresCompatibilities",
                            },
                        ],
                    ),
                    "Service": dict(_service),
                },
            },
            deque(["Resources", "Service", "Properties"]),
            [
                ValidationError(
                    ("'RequiresCompatibilities' is a required property"),
                    validator="required",
                    rule=ServiceFargate(),
                    path_override=deque(["Resources", "TaskDefinition", "Properties"]),
                )
            ],
        ),
    ],
    indirect=["template"],
)
def test_validate(template, start_path, expected, rule, validator):
    for instance, instance_validator in get_value_from_path(
        validator, template, start_path
    ):
        errs = list(rule.validate(instance_validator, "", instance, {}))
        assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest import TestCase

from cfnlint.conditions._rule import _Assertion
from cfnlint.decode import decode_str
from cfnlint.template import Template


class TestConditionsWithRules(TestCase):

    def test_conditions_with_rules(self):
        template = decode_str(
            """
        Conditions:
          IsProd: !Equals [!Ref Environment, "prod"]
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
        Rules:
          Rule1:
            Assertions:
            - Assert:
                Fn::And:
                - !Equals [!Ref Environment, "prod"]
                - !Equals [!Ref "AWS::Region", "us-east-1"]
          Rule2:
            Assertions:
            - Assert:
                Fn::Or:
                - !Equals [!Ref Environment, "prod"]
                - !Equals [!Ref "AWS::Region", "us-east-1"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        self.assertEqual(len(cfn.conditions._rules), 2)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "d0f5e92fc5233a6b011342df171f191838491056",
                "362a2ca660fa34c91feeee4681e8433101d2a687",
            ],
        )

        self.assertTrue(
            cfn.conditions.satisfiable(
                {"IsProd": True, "IsUsEast1": True},
                {"AWS::Region": "us-east-1", "Environment": "prod"},
            )
        )
        self.assertFalse(
            cfn.conditions.satisfiable(
                {"IsProd": True, "IsUsEast1": False},
                {"AWS::Region": "us-west-2", "Environment": "prod"},
            )
        )
        self.assertFalse(
            cfn.conditions.satisfiable(
                {"IsProd": False, "IsUsEast1": True},
                {"AWS::Region": "us-east-1", "Environment": "dev"},
            )
        )
        self.assertFalse(
            cfn.conditions.satisfiable(
                {"IsProd": False, "IsUsEast1": False},
                {"AWS::Region": "us-west-2", "Environment": "dev"},
            )
        )

    def test_conditions_with_rules_implies(self):
        template = decode_str(
            """
        Conditions:
          IsProd: !Equals [!Ref Environment, "prod"]
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
        Rules:
          Rule:
            RuleCondition: !Equals [!Ref Environment, "prod"]
            Assertions:
            - Assert: !Equals [!Ref "AWS::Region", "us-east-1"]

        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        self.assertEqual(len(cfn.conditions._rules), 1)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "d0f5e92fc5233a6b011342df171f191838491056",
                "362a2ca660fa34c91feeee4681e8433101d2a687",
            ],
        )

        self.assertTrue(
            cfn.conditions.satisfiable(
                {"IsProd": True, "IsUsEast1": True},
                {"AWS::Region": "us-east-1", "Environment": "prod"},
            )
        )
        self.assertFalse(
            cfn.conditions.satisfiable(
                {"IsProd": True, "IsUsEast1": False},
                {"AWS::Region": "us-west-2", "Environment": "prod"},
            )
        )
        self.assertTrue(
            cfn.conditions.satisfiable(
                {"IsProd": False, "IsUsEast1": True},
                {"AWS::Region": "us-east-1", "Environment": "dev"},
            )
        )
        self.assertTrue(
            cfn.conditions.satisfiable(
                {"IsProd": False, "IsUsEast1": False},
                {"AWS::Region": "us-west-2", "Environment": "dev"},
            )
        )

    def test_conditions_with_multiple_rules(self):
        template = decode_str(
            """
        Parameters:
          Environment:
            Type: String
            Default: dev
            AllowedValues:
            - dev
            - stage
            - prod
        Conditions:
          IsProd: !Equals [!Ref Environment, "prod"]
          IsDev: !Equals [!Ref Environment, "dev"]
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
          IsNotUsEast1: !Not [!Condition IsUsEast1]
        Rules:
          Rule1:
            RuleCondition: !Equals [!Ref Environment, "prod"]
            Assertions:
            - Assert: !Equals [!Ref "AWS::Region", "us-east-1"]
          Rule2:
            RuleCondition: !Equals [!Ref Environment, "dev"]
            Assertions:
            - Assert: !Not [!Equals [!Ref "AWS::Region", "us-east-1"]]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 4)
        self.assertEqual(len(cfn.conditions._rules), 2)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "d0f5e92fc5233a6b011342df171f191838491056",
                "362a2ca660fa34c91feeee4681e8433101d2a687",
            ],
        )
        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[1].equals],
            [
                "d2dab653475dd270354fe84c4f80b54883e958bb",
                "362a2ca660fa34c91feeee4681e8433101d2a687",
            ],
        )

        self.assertTrue(
            cfn.conditions.satisfiable(
                {
                    "IsProd": True,
                    "IsUsEast1": True,
                    "IsDev": False,
                    "IsNotUsEast1": False,
                },
                {"AWS::Region": "us-east-1", "Environment": "prod"},
            )
        )
        self.assertFalse(
            cfn.conditions.satisfiable(
                {
                    "IsProd": True,
                    "IsUsEast1": False,
                    "IsDev": False,
                    "IsNotUsEast1": False,
                },
                {"AWS::Region": "us-west-2", "Environment": "prod"},
            )
        )
        self.assertFalse(
            cfn.conditions.satisfiable(
                {
                    "IsProd": False,
                    "IsUsEast1": True,
                    "IsDev": True,
                    "IsNotUsEast1": False,
                },
                {"AWS::Region": "us-east-1", "Environment": "dev"},
            )
        )
        self.assertTrue(
            cfn.conditions.satisfiable(
                {
                    "IsProd": False,
                    "IsUsEast1": False,
                    "IsDev": True,
                    "IsNotUsEast1": True,
                },
                {"AWS::Region": "us-west-2", "Environment": "dev"},
            )
        )
        self.assertTrue(
            cfn.conditions.satisfiable(
                {
                    "IsProd": False,
                    "IsUsEast1": True,
                    "IsDev": False,
                    "IsNotUsEast1": False,
                },
                {"AWS::Region": "us-east-1", "Environment": "stage"},
            )
        )
        self.assertTrue(
            cfn.conditions.satisfiable(
                {
                    "IsProd": False,
                    "IsUsEast1": False,
                    "IsDev": False,
                    "IsNotUsEast1": True,
                },
                {"AWS::Region": "us-west-2", "Environment": "stage"},
            )
        )

    def test_fn_equals_assertions_two(self):
        template = decode_str(
            """
        Rules:
          Rule1:
            Assertions:
            - Assert: !Equals ["A", "B"]
          Rule2:
            Assertions:
            - Assert: !Equals ["A", "A"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 0)
        self.assertEqual(len(cfn.conditions._rules), 2)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "e7e68477799682e53ecb09f476128abaeba0bdae",
            ],
        )
        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[1].equals],
            [
                "da2a95009a205d5caacd42c3c11ebd4c151b3409",
            ],
        )

        self.assertFalse(
            cfn.conditions.satisfiable(
                {},
                {},
            )
        )

    def test_fn_equals_assertions_one(self):
        template = decode_str(
            """
        Rules:
          Rule1:
            Assertions:
            - Assert: !Equals ["A", "A"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 0)
        self.assertEqual(len(cfn.conditions._rules), 1)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "da2a95009a205d5caacd42c3c11ebd4c151b3409",
            ],
        )

        self.assertTrue(
            cfn.conditions.satisfiable(
                {},
                {},
            )
        )

    def test_fn_equals_assertions_ref_no_data(self):
        template = decode_str(
            """
        Parameters:
            AccountId:
                Type: String
        Rules:
          Rule1:
            Assertions:
            - Assert: !Equals [!Ref AccountId, !Ref AWS::AccountId]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 0)
        self.assertEqual(len(cfn.conditions._rules), 1)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "f36e61f3d5bf6cdc6ea2e7f01487af728094a439",
            ],
        )

        self.assertTrue(
            cfn.conditions.satisfiable(
                {},
                {},
            )
        )

    def test_fn_equals_assertions_ref_never_satisfiable(self):
        template = decode_str(
            """
        Parameters:
            AccountId:
                Type: String
        Rules:
          Rule1:
            Assertions:
            - Assert: !Equals [!Ref AccountId, !Ref AWS::AccountId]
            - Assert: !Not [!Equals [!Ref AccountId, !Ref AWS::AccountId]]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 0)
        self.assertEqual(len(cfn.conditions._rules), 1)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "f36e61f3d5bf6cdc6ea2e7f01487af728094a439",
                "f36e61f3d5bf6cdc6ea2e7f01487af728094a439",
            ],
        )

        self.assertFalse(
            cfn.conditions.satisfiable(
                {},
                {},
            )
        )

    def test_conditions_with_rules_and_parameters(self):
        template = decode_str(
            """
        Conditions:
            DeployGateway: !Equals
                - !Ref 'DeployGateway'
                - 'true'
            DeployVpc: !Equals
                - !Ref 'DeployVpc'
                - 'true'
        Parameters:
            DeployAnything:
                AllowedValues:
                - 'false'
                - 'true'
                Type: 'String'
            DeployGateway:
                AllowedValues:
                - 'false'
                - 'true'
                Type: 'String'
            DeployVpc:
                AllowedValues:
                - 'false'
                - 'true'
                Type: 'String'
        Rules:
            DeployGateway:
                Assertions:
                - Assert: !Or
                    - !Equals
                        - !Ref 'DeployAnything'
                        - 'true'
                    - !Equals
                        - !Ref 'DeployGateway'
                        - 'false'
            DeployVpc:
                Assertions:
                - Assert: !Or
                    - !Equals
                        - !Ref 'DeployGateway'
                        - 'true'
                    - !Equals
                        - !Ref 'DeployVpc'
                        - 'false'
        Resources:
            InternetGateway:
                Condition: 'DeployGateway'
                Type: 'AWS::EC2::InternetGateway'
            InternetGatewayAttachment:
                Condition: 'DeployVpc'
                Type: 'AWS::EC2::VPCGatewayAttachment'
                Properties:
                    InternetGatewayId: !Ref 'InternetGateway'
                    VpcId: !Ref 'Vpc'
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._conditions), 2)
        self.assertEqual(len(cfn.conditions._rules), 2)

        self.assertListEqual(
            [equal.hash for equal in cfn.conditions._rules[0].equals],
            [
                "d0d70a1e66dc83d7a0fce24c2eca396af1f34e53",
                "bbf5c94c1a4b5a79c7a7863fe9463884cb422450",
            ],
        )

        self.assertTrue(
            cfn.conditions.satisfiable(
                {},
                {},
            )
        )

        self.assertTrue(
            cfn.conditions.check_implies({"DeployVpc": True}, "DeployGateway")
        )

        self.assertFalse(
            cfn.conditions.check_implies({"DeployVpc": False}, "DeployGateway")
        )

        self.assertFalse(
            cfn.conditions.check_implies({"DeployGateway": False}, "DeployVpc")
        )


class TestAssertion(TestCase):
    def test_assertion_errors(self):
        with self.assertRaises(ValueError):
            _Assertion({"A": "B", "C": "D"}, {})

        with self.assertRaises(ValueError):
            _Assertion({"Fn::Not": {"C": "D"}}, {})

        with self.assertRaises(ValueError):
            _Assertion({"Not": {"C": "D"}}, {})

        with self.assertRaises(ValueError):
            _Assertion({"Condition": {"C": "D"}}, {})

    def test_init_rules_with_list(self):
        template = decode_str(
            """
        Conditions:
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
          IsNotUsEast1: !Not [!Condition IsUsEast1]
        Rules: []
        """
        )[0]

        cfn = Template("", template)
        self.assertListEqual(cfn.conditions._rules, [])

    def test_init_rules_with_wrong_assertions_type(self):
        template = decode_str(
            """
        Conditions:
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
          IsNotUsEast1: !Not [!Condition IsUsEast1]
        Rules:
          Rule1:
            Assertions: {"Foo": "Bar"}
          Rule2:
            Assertions:
            - Assert: !Equals [!Ref "AWS::Region", "us-east-1"]
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._rules), 1)

    def test_init_rules_with_no_keys(self):
        template = decode_str(
            """
        Conditions:
          IsUsEast1: !Equals [!Ref "AWS::Region", "us-east-1"]
          IsNotUsEast1: !Not [!Condition IsUsEast1]
        Rules:
          Rule1:
            Foo: Bar
          Rule2:
            Assertions:
            - Assert:
                Fn::Or:
                - !Not [!Equals [!Ref "AWS::Region", "us-east-1"]]
                - !Equals [!Ref "AWS::Region", "us-east-1"]
          Rule3: []
        """
        )[0]

        cfn = Template("", template)
        self.assertEqual(len(cfn.conditions._rules), 1)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import json
import logging
from test.testlib.testcase import BaseTestCase
from unittest.mock import patch

import cfnlint.maintenance

LOGGER = logging.getLogger("cfnlint.maintenance")
LOGGER.addHandler(logging.NullHandler())


class TestUpdateIamPolicies(BaseTestCase):
    """Used for Testing Rules"""

    @patch("cfnlint.maintenance.get_url_content")
    @patch("cfnlint.maintenance.json.dump")
    def test_update_iam_policies(self, mock_json_dump, mock_content):
        """Success update iam policies"""

        services = [
            {
                "service": "foo",
                "url": "https://servicereference.us-east-1.amazonaws.com/v1/foo/foo.json",
            }
        ]

        cloudformation = {
            "Name": "cloudformation",
            "Actions": [
                {
                    "Name": "CreateStack",
                    "ActionConditionKeys": [
                        "aws:RequestTag/${TagKey}",
                        "aws:TagKeys",
                        "cloudformation:ResourceTypes",
                        "cloudformation:RoleArn",
                        "cloudformation:StackPolicyUrl",
                        "cloudformation:TemplateUrl",
                    ],
                    "Resources": [{"Name": "stack"}],
                }
            ],
            "Resources": [
                {
                    "Name": "stack",
                    "ARNFormats": [
                        "arn:${Partition}:cloudformation:${Region}:${Account}:stack/${StackName}/${Id}"
                    ],
                    "ConditionKeys": ["aws:ResourceTag/${TagKey}"],
                },
                {
                    "Name": "stackset",
                    "ARNFormats": [
                        "arn:${Partition}:cloudformation:${Region}:${Account}:${StackSetName}"
                    ],
                },
            ],
        }

        mock_content.side_effect = [
            json.dumps(services),
            json.dumps(cloudformation),
        ]

        builtin_module_name = "builtins"

        with patch("{}.open".format(builtin_module_name)) as mock_builtin_open:
            cfnlint.maintenance.update_iam_policies()
            mock_json_dump.assert_called_with(
                {
                    "foo": {
                        "Actions": {"createstack": {"Resources": ["stack"]}},
                        "Resources": {
                            "stack": {
                                "ARNFormats": [
                                    "arn:${Partition}:cloudformation:${Region}:${Account}:stack/.*"
                                ],
                                "ConditionKeys": ["aws:ResourceTag/${TagKey}"],
                            },
                            "stackset": {
                                "ARNFormats": [
                                    "arn:${Partition}:cloudformation:${Region}:${Account}:.*"
                                ]
                            },
                        },
                    }
                },
                mock_builtin_open.return_value.__enter__.return_value,
                indent=1,
                separators=(",", ": "),
                sort_keys=True,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from unittest.mock import patch

import pytest

from cfnlint.context.context import Resource, _init_resources
from cfnlint.match import Match
from cfnlint.rules.errors.parse import ParseError
from cfnlint.schema import AttributeDict


@pytest.mark.parametrize(
    "name,instance,expected_ref",
    [
        (
            "Valid resource",
            {"Type": "AWS::EC2::VPC"},
            {"format": "AWS::EC2::VPC.Id", "type": "string"},
        ),
    ],
)
def test_resource(name, instance, expected_ref):
    region = "us-east-1"
    resource = Resource(instance)

    assert expected_ref == resource.ref(
        region
    ), f"{name!r} test got {resource.ref(region)}"


@pytest.mark.parametrize(
    "name,instance",
    [
        ("Invalid Type", {"Type": {}}),
    ],
)
def test_errors(name, instance):
    with pytest.raises(ValueError):
        Resource(instance)


def test_resources():
    with pytest.raises(ValueError):
        _init_resources([])


@pytest.mark.parametrize(
    "name,instance,filename,decode_results,expected_getatts",
    [
        (
            "Nested stack with no Properties",
            {"Type": "AWS::CloudFormation::Stack"},
            "foo/bar.yaml",
            None,
            AttributeDict({"Outputs\\..*": "/properties/CfnLintStringType"}),
        ),
        (
            "Nested stack with template URL",
            {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {"TemplateURL": "https://bucket/path.yaml"},
            },
            "foo/bar.yaml",
            None,
            AttributeDict({"Outputs\\..*": "/properties/CfnLintStringType"}),
        ),
        (
            "Nested stack with a None filename",
            {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {"TemplateURL": "application.yaml"},
            },
            None,
            None,
            AttributeDict({"Outputs\\..*": "/properties/CfnLintStringType"}),
        ),
        (
            "Nested stack with a local file",
            {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {"TemplateURL": "./bar.yaml"},
            },
            "foo/bar.yaml",
            ({"Outputs": {"MyValue": {"Type": "String"}}}, None),
            AttributeDict({"Outputs.MyValue": "/properties/CfnLintStringType"}),
        ),
        (
            "Nested stack with a local file and no outputs",
            {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {"TemplateURL": "./bar.yaml"},
            },
            "foo/bar.yaml",
            ({}, None),
            AttributeDict({}),
        ),
        (
            "Nested stack with a local file but match error",
            {
                "Type": "AWS::CloudFormation::Stack",
                "Properties": {"TemplateURL": "./bar.yaml"},
            },
            "foo/bar.yaml",
            (None, Match("test", rule=ParseError())),
            AttributeDict({"Outputs\\..*": "/properties/CfnLintStringType"}),
        ),
    ],
)
def test_nested_stacks(name, instance, filename, decode_results, expected_getatts):
    region = "us-east-1"

    with patch("cfnlint.decode.decode") as mock_decode:
        if decode_results is not None:
            mock_decode.return_value = decode_results

        resource = Resource(instance, filename)

        assert expected_getatts == resource.get_atts(
            region
        ), f"{name!r} test got {resource.get_atts(region)}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque

import jsonpatch
import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.resources.ecs.TaskDefinitionAwsVpc import TaskDefinitionAwsVpc


@pytest.fixture
def rule():
    rule = TaskDefinitionAwsVpc()
    yield rule


_task_definition = {
    "Type": "AWS::ECS::TaskDefinition",
    "Properties": {
        "NetworkMode": "awsvpc",
        "ContainerDefinitions": [
            {
                "Name": "my-container",
                "Image": "my-image",
                "PortMappings": [
                    {
                        "ContainerPort": 8080,
                    }
                ],
            }
        ],
    },
}


@pytest.mark.parametrize(
    "template,start_path,expected",
    [
        (
            {
                "Resources": {
                    "TaskDefinition": dict(_task_definition),
                }
            },
            deque(["Resources", "TaskDefinition", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": (
                                    "/Properties/ContainerDefinitions/"
                                    "0/PortMappings/0/HostPort"
                                ),
                                "value": "8080",
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "TaskDefinition", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": (
                                    "/Properties/ContainerDefinitions/"
                                    "0/PortMappings/0/HostPort"
                                ),
                                "value": "8080",
                            },
                            {
                                "op": "replace",
                                "path": (
                                    "/Properties/ContainerDefinitions/"
                                    "0/PortMappings/0/ContainerPort"
                                ),
                                "value": {"Fn::Sub": "8080"},
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "TaskDefinition", "Properties"]),
            [],
        ),
        (
            {
                "Parameters": {
                    "MyPort": {"Type": "String"},
                    "MySecondPort": {"Type": "String"},
                },
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": (
                                    "/Properties/ContainerDefinitions/0"
                                    "/PortMappings/0/ContainerPort"
                                ),
                                "value": {"Ref": "MyPort"},
                            },
                            {
                                "op": "add",
                                "path": (
                                    "/Properties/ContainerDefinitions/0"
                                    "/PortMappings/0/HostPort"
                                ),
                                "value": {"Ref": "MySecondPort"},
                            },
                        ],
                    ),
                },
            },
            deque(["Resources", "TaskDefinition", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": (
                                    "/Properties/ContainerDefinitions/0"
                                    "/PortMappings/0/HostPort"
                                ),
                                "value": "80",
                            },
                            {
                                "op": "replace",
                                "path": ("/Properties/NetworkMode"),
                                "value": "bridge",
                            },
                        ],
                    ),
                }
            },
            deque(["Resources", "TaskDefinition", "Properties"]),
            [],
        ),
        (
            {
                "Resources": {
                    "TaskDefinition": jsonpatch.apply_patch(
                        dict(_task_definition),
                        [
                            {
                                "op": "add",
                                "path": (
                                    "/Properties/ContainerDefinitions/0"
                                    "/PortMappings/0/HostPort"
                                ),
                                "value": "80",
                            }
                        ],
                    ),
                }
            },
            deque(["Resources", "TaskDefinition", "Properties"]),
            [
                ValidationError(
                    ("'80' does not equal 8080"),
                    validator="const",
                    rule=TaskDefinitionAwsVpc(),
                    path_override=deque(
                        [
                            "Resources",
                            "TaskDefinition",
                            "Properties",
                            "ContainerDefinitions",
                            0,
                            "PortMappings",
                            0,
                            "HostPort",
                        ]
                    ),
                )
            ],
        ),
    ],
    indirect=["template"],
)
def test_validate(template, start_path, expected, rule, validator):
    for instance, instance_validator in get_value_from_path(
        validator, template, start_path
    ):
        errs = list(rule.validate(instance_validator, "", instance, {}))

        assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest.mock import patch

import pytest

from cfnlint.jsonschema.exceptions import ValidationError
from cfnlint.jsonschema.validators import CfnTemplateValidator


def fail(validator, errors, instance, schema):
    for each in errors:
        each.setdefault("message", "You told me to fail!")
        yield ValidationError(**each)


@pytest.fixture
def validator():
    validators = {"fail": fail}
    validator = CfnTemplateValidator({}).extend(
        validators=validators,
    )({})
    return validator


def _eq(self, other):
    return self.message == other.message


def _lt(self, other):
    return hash(self.message) < hash(other.message)


@pytest.mark.parametrize(
    "name,schema,instance,expected",
    [
        ("Success", {"fail": []}, "hello", []),
        (
            "One Error",
            {"fail": [{"message": "Whoops!"}]},
            "goodbye",
            [
                ValidationError(
                    "Whoops!",
                    instance="goodbye",
                    schema={"fail": [{"message": "Whoops!"}]},
                    validator="fail",
                    validator_value=[{"message": "Whoops!"}],
                    schema_path=deque(["fail"]),
                )
            ],
        ),
        (
            "Multiple Errors",
            {
                "fail": [
                    {"message": "First"},
                    {"message": "Second!", "validator": "asdf"},
                    {"message": "Third"},
                ]
            },
            "goodbye",
            [
                ValidationError(
                    "First",
                    instance="goodbye",
                    schema={"message": "First"},
                    validator="fail",
                    validator_value=[{"message": "Whoops!"}],
                    schema_path=deque(["fail"]),
                ),
                ValidationError(
                    "Second!",
                    instance="goodbye",
                    schema={"message": "Second!", "validator": "asdf"},
                    validator="asdf",
                    validator_value=[{"message": "Whoops!"}],
                    schema_path=deque(["fail"]),
                ),
                ValidationError(
                    "Third",
                    instance="goodbye",
                    schema={"message": "Third"},
                    validator="fail",
                    validator_value=[{"message": "Whoops!"}],
                    schema_path=deque(["fail"]),
                ),
            ],
        ),
    ],
)
def test_validator(name, schema, instance, expected, validator):
    validator = validator.evolve(schema=schema)
    errs = list(validator.iter_errors(instance))
    assert errs == expected, f"{name!r} returned {errs!r}"


@pytest.mark.parametrize(
    "name,schema,instance,expected",
    [
        (
            "Valid type single",
            {"type": "string"},
            "foo",
            [],
        ),
        (
            "Valid type single array",
            {"type": ["string"]},
            "foo",
            [],
        ),
        (
            "Valid type many in array",
            {"type": ["string", "object"]},
            "foo",
            [],
        ),
        (
            "Single type failure",
            {"type": "string"},
            [],
            [ValidationError("[] is not of type 'string'")],
        ),
        (
            "Multiple type failure",
            {"type": ["string", "object"]},
            [],
            [
                ValidationError(
                    "[] is not of type 'string', 'object'",
                )
            ],
        ),
        (
            "Multiple type failure",
            {"type": "string"},
            [],
            [ValidationError("[] is not of type 'string'")],
        ),
        (
            "valid minimum",
            {"minimum": 2},
            2,
            [],
        ),
        (
            "valid minimum with string",
            {"minimum": 2},
            "2",
            [],
        ),
        (
            "valid minimum with wrong type as string",
            {"minimum": 2},
            "a",
            [],
        ),
        (
            "valid minimum with wrong type",
            {"minimum": 2},
            {},
            [],
        ),
        (
            "valid minimum with wrong type boolean",
            {"minimum": 2},
            True,
            [],
        ),
        (
            "minimum",
            {"minimum": 2},
            1,
            [ValidationError("1 is less than the minimum of 2")],
        ),
        (
            "minimum with string",
            {"minimum": 2},
            "1",
            [
                ValidationError(
                    "'1' is less than the minimum of 2",
                )
            ],
        ),
        (
            "valid maximum",
            {"maximum": 2},
            "2",
            [],
        ),
        (
            "valid maximum with string",
            {"minimum": 2},
            "2",
            [],
        ),
        (
            "valid maximum with wrong type",
            {"maximum": 2},
            {},
            [],
        ),
        (
            "valid maximum with wrong type boolean",
            {"maximum": 2},
            True,
            [],
        ),
        (
            "valid maximum with wrong type as string",
            {"maximum": 2},
            "a",
            [],
        ),
        (
            "maximum",
            {"maximum": 0},
            1,
            [ValidationError("1 is greater than the maximum of 0")],
        ),
        (
            "maximum with string",
            {"maximum": 0},
            "1",
            [
                ValidationError(
                    "'1' is greater than the maximum of 0",
                )
            ],
        ),
        (
            "valid dependencies wrong type",
            {"dependencies": {"bar": ["foo"]}},
            [],
            [],
        ),
        (
            "dependencies list",
            {"dependencies": {"bar": ["foo"]}},
            {"bar": 2},
            [
                ValidationError(
                    "'foo' is a dependency of 'bar'",
                )
            ],
        ),
        (
            "valid dependencies when property not included",
            {"dependencies": {"bar": ["foo"]}},
            {"foo": 2},
            [],
        ),
        (
            "dependencies object",
            {"dependencies": {"foo": {"required": ["bar"]}}},
            {"foo": 2},
            [
                ValidationError(
                    "'bar' is a required property",
                )
            ],
        ),
        (
            "valid additionalProperties",
            {"additionalProperties": True},
            {"foo": "bar"},
            [],
        ),
        (
            "valid additionalProperties with wrong type",
            {"additionalProperties": False},
            [],
            [],
        ),
        (
            "additionalProperties false",
            {"additionalProperties": False},
            {"foo": 2},
            [
                ValidationError(
                    "Additional properties are not allowed ('foo' was unexpected)",
                )
            ],
        ),
        (
            "additionalProperties false",
            {"additionalProperties": False, "properties": {"foooooo": True}},
            {"foooooa": 2},
            [
                ValidationError(
                    (
                        "Additional properties are not allowed ('foooooa' "
                        "was unexpected. Did you mean 'foooooo'?)"
                    ),
                )
            ],
        ),
        (
            "additionalProperties object",
            {"additionalProperties": {"type": "string"}},
            {"foo": []},
            [
                ValidationError(
                    "[] is not of type 'string'",
                )
            ],
        ),
        (
            "additionalProperties with patternProperties",
            {"additionalProperties": False, "patternProperties": {"^bar$": False}},
            {"foo": []},
            [
                ValidationError(
                    "'foo' does not match any of the regexes: '^bar$'",
                )
            ],
        ),
        (
            "additionalProperties with multiple properties",
            {"additionalProperties": False},
            {"foo": [], "bar": []},
            [
                ValidationError(
                    "Additional properties are not allowed ('foo' was unexpected)",
                ),
                ValidationError(
                    "Additional properties are not allowed ('bar' was unexpected)",
                ),
            ],
        ),
        (
            "valid const",
            {"const": 12},
            12,
            [],
        ),
        (
            "const",
            {"const": 12},
            {"foo": "bar"},
            [
                ValidationError(
                    "12 was expected",
                )
            ],
        ),
        (
            "valid enum",
            {"enum": ["bar"]},
            "bar",
            [],
        ),
        (
            "enum",
            {"enum": ["bar"]},
            "foo",
            [
                ValidationError(
                    "'foo' is not one of ['bar']",
                    validator="enum",
                    schema_path=deque(["enum"]),
                )
            ],
        ),
        (
            "valid enum int",
            {"enum": [0]},
            "0",
            [],
        ),
        (
            "enum int",
            {"enum": [0]},
            1,
            [ValidationError("1 is not one of [0]")],
        ),
        (
            "format default message",
            {"format": "date-time"},
            "bla",
            [
                ValidationError(
                    "'bla' is not a 'date-time'",
                )
            ],
        ),
        (
            "a false schema",
            False,
            "foo",
            [ValidationError("False schema does not allow 'foo'")],
        ),
        (
            "valid if",
            {"if": {"type": "string"}, "then": True, "else": False},
            "",
            [],
        ),
        (
            "if then with false",
            {"if": {"type": "string"}, "then": False, "else": True},
            "a",
            [
                ValidationError(
                    "False schema does not allow 'a'",
                )
            ],
        ),
        (
            "valid else",
            {"if": {"type": "string"}, "then": False, "else": True},
            [],
            [],
        ),
        (
            "if then else with false",
            {"if": {"type": "string"}, "then": True, "else": False},
            [],
            [
                ValidationError(
                    "False schema does not allow []",
                )
            ],
        ),
        (
            "valid items",
            {"items": {"type": "string"}},
            ["foo"],
            [],
        ),
        (
            "valid items with wrong type",
            {"items": {"type": "string"}},
            {},
            [],
        ),
        (
            "items object",
            {"items": {"type": "string"}},
            [[]],
            [
                ValidationError(
                    "[] is not of type 'string'",
                )
            ],
        ),
        (
            "items with prefixItems",
            {"items": {"const": "2"}, "prefixItems": [{"const": "1"}]},
            ["1", "1"],
            [
                ValidationError(
                    "'2' was expected",
                )
            ],
        ),
        (
            "items with prefixItems and a validation error",
            {"items": {"const": "2"}, "prefixItems": [{"const": "1"}]},
            ["1", "2"],
            [],
        ),
        (
            "prefixItems ",
            {"prefixItems": [{"type": "string"}]},
            [[]],
            [
                ValidationError(
                    "[] is not of type 'string'",
                )
            ],
        ),
        (
            "valid multipleOf",
            {"multipleOf": 2},
            2,
            [],
        ),
        (
            "valid multipleOf with string",
            {"multipleOf": 2},
            "2",
            [],
        ),
        (
            "valid multipleOf with wrong type",
            {"multipleOf": 2},
            [],
            [],
        ),
        (
            "valid multipleOf with wrong type string",
            {"multipleOf": 2},
            "A",
            [],
        ),
        (
            "multipleOf",
            {"multipleOf": 2},
            3,
            [
                ValidationError(
                    "3 is not a multiple of 2",
                )
            ],
        ),
        (
            "multipleOf with string",
            {"multipleOf": 2},
            "3",
            [
                ValidationError(
                    "'3' is not a multiple of 2",
                )
            ],
        ),
        (
            "multipleOf with number",
            {"multipleOf": 2.0},
            "3",
            [
                ValidationError(
                    "'3' is not a multiple of 2.0",
                )
            ],
        ),
        (
            "valid minItems",
            {"minItems": 1},
            ["foo"],
            [],
        ),
        (
            "valid minItems with wrong type",
            {"minItems": 1},
            {},
            [],
        ),
        (
            "minItems",
            {"minItems": 2},
            [],
            [
                ValidationError(
                    "expected minimum item count: 2, found: 0",
                )
            ],
        ),
        (
            "valid maxItems",
            {"maxItems": 1},
            ["foo"],
            [],
        ),
        (
            "valid maxItems with wrong type",
            {"maxItems": 1},
            {},
            [],
        ),
        (
            "maxItems",
            {"maxItems": 0},
            ["foo"],
            [ValidationError("expected maximum item count: 0, found: 1")],
        ),
        (
            "valid minLength",
            {"minLength": 1},
            "foo",
            [],
        ),
        (
            "valid minLength with wrong type",
            {"minLength": 1},
            1,
            [],
        ),
        (
            "minLength",
            {"minLength": 2},
            "",
            [
                ValidationError(
                    "expected minimum length: 2, found: 0",
                )
            ],
        ),
        (
            "valid maxLength",
            {"maxLength": 3},
            "foo",
            [],
        ),
        (
            "valid maxLength with wrong type",
            {"maxLength": 1},
            1,
            [],
        ),
        (
            "maxLength",
            {"maxLength": 0},
            "foo",
            [
                ValidationError(
                    "expected maximum length: 0, found: 3",
                )
            ],
        ),
        (
            "valid not",
            {"not": False},
            True,
            [],
        ),
        (
            "not",
            {"not": True},
            False,
            [
                ValidationError(
                    "False should not be valid under True",
                )
            ],
        ),
        (
            "valid pattern",
            {"pattern": "^foo$"},
            "foo",
            [],
        ),
        (
            "valid pattern with wrong type",
            {"pattern": "^foo$"},
            {},
            [],
        ),
        (
            "pattern",
            {"pattern": "^a*$"},
            "bbb",
            [
                ValidationError(
                    "'bbb' does not match '^a*$'",
                )
            ],
        ),
        (
            "valid contains",
            {"contains": {"type": "string"}},
            ["foo"],
            [],
        ),
        (
            "valid contains with min and max",
            {"contains": {"type": "string"}, "minContains": 2, "maxContains": 2},
            ["foo", "bar", 2],
            [],
        ),
        (
            "valid contains with wrong type",
            {"contains": {"type": "string"}},
            {},
            [],
        ),
        (
            "contains",
            {"contains": {"type": "string"}},
            [],
            [
                ValidationError(
                    "[] does not contain items matching the given schema",
                )
            ],
        ),
        (
            "invalid contains with min",
            {"contains": {"type": "string"}, "minContains": 2},
            ["foo", 2],
            [
                ValidationError(
                    (
                        "Too few items match the given schema "
                        "(expected at least 2 but only 1 matched)"
                    ),
                    validator="minContains",
                    validator_value=2,
                )
            ],
        ),
        (
            "invalid contains with min and empty match",
            {"contains": {"type": "string"}, "minContains": 2},
            [2],
            [
                ValidationError(
                    "[2] does not contain items matching the given schema",
                )
            ],
        ),
        (
            "invalid contains with  max",
            {"contains": {"type": "string"}, "maxContains": 2},
            ["foo", "bar", "foobar", 2],
            [
                ValidationError(
                    "Too many items match the given schema (expected at most 2)",
                    validator="maxContains",
                    validator_value=2,
                )
            ],
        ),
        (
            "valid exclusiveMinimum",
            {"exclusiveMinimum": 4},
            5,
            [],
        ),
        (
            "valid exclusiveMinimum with wrong type",
            {"exclusiveMinimum": 4},
            {},
            [],
        ),
        (
            "valid exclusiveMinimum with non number string",
            {"exclusiveMinimum": 4},
            "foo",
            [],
        ),
        (
            "exclusiveMinimum",
            {"exclusiveMinimum": 5},
            5,
            [
                ValidationError(
                    "5 is less than or equal to the minimum of 5",
                )
            ],
        ),
        (
            "exclusiveMinimum with string",
            {"exclusiveMinimum": 5},
            "5",
            [
                ValidationError(
                    "'5' is less than or equal to the minimum of 5",
                )
            ],
        ),
        (
            "valid exclusiveMaximum",
            {"exclusiveMaximum": 5},
            4,
            [],
        ),
        (
            "valid exclusiveMaximum with wrong type",
            {"exclusiveMaximum": 5},
            {},
            [],
        ),
        (
            "valid exclusiveMaximum with wrong string type",
            {"exclusiveMaximum": 5},
            "foo",
            [],
        ),
        (
            "exclusiveMaximum with string",
            {"exclusiveMaximum": 5},
            "5",
            [
                ValidationError(
                    "'5' is greater than or equal to the maximum of 5",
                )
            ],
        ),
        (
            "valid required",
            {"required": ["foo"]},
            {"foo": False},
            [],
        ),
        (
            "valid required wrong type",
            {"required": ["foo"]},
            [],
            [],
        ),
        (
            "required",
            {"required": ["foo"]},
            {},
            [
                ValidationError(
                    "'foo' is a required property",
                )
            ],
        ),
        (
            "valid minProperties",
            {"minProperties": 1},
            {"foo": False},
            [],
        ),
        (
            "valid minProperties with wrong type",
            {"minProperties": 1},
            [],
            [],
        ),
        (
            "minProperties",
            {"minProperties": 1},
            {},
            [
                ValidationError(
                    "expected minimum property count: 1, found: 0",
                )
            ],
        ),
        (
            "valid maxProperties",
            {"maxProperties": 1},
            {"foo": False},
            [],
        ),
        (
            "valid maxProperties with wrong type",
            {"maxProperties": 1},
            [],
            [],
        ),
        (
            "maxProperties",
            {"maxProperties": 1},
            {"foo": {}, "bar": {}},
            [
                ValidationError(
                    "expected maximum property count: 1, found: 2",
                )
            ],
        ),
        (
            "valid patternProperties",
            {"patternProperties": {"^foo$": True}},
            {"foo": {}},
            [],
        ),
        (
            "valid patternProperties with wrong type",
            {"patternProperties": {"^foo$": True}},
            [],
            [],
        ),
        (
            "patternProperties",
            {"patternProperties": {"^foo$": False}},
            {"foo": "bar"},
            [
                ValidationError(
                    "False schema does not allow 'bar'",
                )
            ],
        ),
        (
            "valid properties",
            {"properties": {"foo": True}},
            {"foo": "bar"},
            [],
        ),
        (
            "valid properties with wrong type",
            {"properties": {"foo": True}},
            [],
            [],
        ),
        (
            "properties",
            {"properties": {"foo": False}},
            {"foo": "bar"},
            [
                ValidationError(
                    "False schema does not allow 'bar'",
                )
            ],
        ),
        (
            "valid propertyNames",
            {"propertyNames": {"pattern": "^bar$"}},
            {"bar": "foo"},
            [],
        ),
        (
            "valid propertyNames with wrong type",
            {"propertyNames": {"pattern": "^bar$"}},
            [],
            [],
        ),
        (
            "propertyNames",
            {"propertyNames": {"pattern": "^bar$"}},
            {"foo": "bar"},
            [
                ValidationError(
                    "'foo' does not match '^bar$'",
                )
            ],
        ),
        (
            "valid oneOf matches none",
            {"oneOf": [True]},
            False,
            [],
        ),
        (
            "oneOf matches none",
            {"oneOf": [False]},
            {},
            [
                ValidationError(
                    "{} is not valid under any of the given schemas",
                )
            ],
        ),
        (
            "oneOf matches too many",
            {"oneOf": [True, True]},
            {},
            [
                ValidationError(
                    "{} is valid under each of True, True",
                )
            ],
        ),
        (
            "valid allOf",
            {"allOf": [True, True]},
            False,
            [],
        ),
        (
            "allOf matches one False",
            {"allOf": [True, False]},
            {},
            [
                ValidationError(
                    "False schema does not allow {}",
                )
            ],
        ),
        (
            "allOf matches multiple False",
            {"allOf": [True, False, False]},
            {},
            [
                ValidationError(
                    "False schema does not allow {}",
                ),
                ValidationError(
                    "False schema does not allow {}",
                ),
            ],
        ),
        (
            "valid anyOf matches more than one",
            {"anyOf": [True, True]},
            False,
            [],
        ),
        (
            "anyOf matches none",
            {"anyOf": [False, False]},
            {},
            [
                ValidationError(
                    "{} is not valid under any of the given schemas",
                )
            ],
        ),
        (
            "valid uniqueItems",
            {"uniqueItems": True},
            [1, 2, "3"],
            [],
        ),
        (
            "uniqueItems",
            {"uniqueItems": True},
            [1, 2, "1"],
            [
                ValidationError(
                    "array items are not unique",
                )
            ],
        ),
        (
            "valid requiredOr",
            {"requiredOr": ["foo", "bar"]},
            {"foo": {}},
            [],
        ),
        (
            "valid requiredOr with wrong type",
            {"requiredOr": ["foo", "bar"]},
            [],
            [],
        ),
        (
            "invalid requiredOr with empty object",
            {"requiredOr": ["foo", "bar"]},
            {},
            [
                ValidationError(
                    "One of ['foo', 'bar'] is a required property",
                )
            ],
        ),
        (
            "valid requiredXor",
            {"requiredXor": ["foo", "bar"]},
            {"foo": {}},
            [],
        ),
        (
            "valid requiredXor with wrong type",
            {"requiredXor": ["foo", "bar"]},
            [],
            [],
        ),
        (
            "requiredXor",
            {"requiredXor": ["foo", "bar"]},
            {},
            [
                ValidationError(
                    "Only one of ['foo', 'bar'] is a required property",
                )
            ],
        ),
        (
            "requiredXor with multiple errors",
            {"requiredXor": ["foo", "bar"]},
            {"foo": "foo", "bar": "bar"},
            [
                ValidationError(
                    "Only one of ['foo', 'bar'] is a required property",
                ),
                ValidationError(
                    "Only one of ['foo', 'bar'] is a required property",
                ),
            ],
        ),
        (
            "valid uniqueKeys",
            {"uniqueKeys": ["Name"]},
            [{"Name": "foo"}, {"Name": "bar"}],
            [],
        ),
        (
            "valid uniqueKeys with wrong types",
            {"uniqueKeys": ["Name"]},
            {"Name": "foo"},
            [],
        ),
        (
            "valid uniqueKeys with wrong type in array",
            {"uniqueKeys": ["Name"]},
            [{"Name": "foo"}, []],
            [],
        ),
        (
            "uniqueKeys",
            {"uniqueKeys": ["Name"]},
            [
                {
                    "Name": "foo",
                },
                {
                    "Name": "foo",
                },
            ],
            [
                ValidationError(
                    "array items are not unique " "for keys ['Name']",
                )
            ],
        ),
        (
            "valid dependentExcluded when wrong type",
            {"dependentExcluded": {"foo": ["bar"]}},
            [],
            [],
        ),
        (
            "valid dependentExcluded when not specified",
            {"dependentExcluded": {"foo": ["bar"]}},
            {"bar": "bar"},
            [],
        ),
        (
            "dependentExcluded",
            {"dependentExcluded": {"foo": ["bar"], "bar": ["foo"]}},
            {"foo": "foo", "bar": "bar"},
            [
                ValidationError("'bar' should not be included with 'foo'"),
                ValidationError("'foo' should not be included with 'bar'"),
            ],
        ),
        (
            "dependentRequired",
            {"dependentRequired": {"foo": ["bar"]}},
            {"foo": "foo"},
            [
                ValidationError("'bar' is a dependency of 'foo'"),
            ],
        ),
        (
            "dependentRequired with multiple properties",
            {"dependentRequired": {"foo": ["a", "b"]}},
            {"foo": "foo"},
            [
                ValidationError("'a' is a dependency of 'foo'"),
                ValidationError("'b' is a dependency of 'foo'"),
            ],
        ),
        (
            "valid dependentRequired when property not included",
            {"dependentRequired": {"foo": ["a", "b"]}},
            {"bar": "bar"},
            [],
        ),
        (
            "valid dependentRequired",
            {"dependentRequired": {"foo": ["bar"]}},
            {"foo": "foo", "bar": "bar"},
            [],
        ),
        (
            "valid dependentRequired with wrong type",
            {"dependentRequired": {"foo": ["bar"]}},
            [],
            [],
        ),
        (
            "valid prefixItems",
            {"prefixItems": [{"type": "string"}]},
            ["foo"],
            [],
        ),
        (
            "invalid prefixItems with wrong type",
            {"prefixItems": [{"type": "string"}]},
            [1],
            [
                ValidationError("1 is not of type 'string'"),
            ],
        ),
        (
            "valid enum case insensitive",
            {"enumCaseInsensitive": ["A"]},
            "a",
            [],
        ),
        (
            "valid enum case insensitive with non string",
            {"enumCaseInsensitive": ["a", 1]},
            "A",
            [],
        ),
        (
            "valid enum case insensitive with non string value",
            {"enumCaseInsensitive": ["A", 1]},
            1,
            [],
        ),
        (
            "valid enum case insensitive with non string value",
            {"enumCaseInsensitive": ["A"]},
            "b",
            [
                ValidationError("'b' is not one of ['a'] (case-insensitive)"),
            ],
        ),
    ],
)
@patch.object(ValidationError, "__eq__", spec=True, new=_eq)
@patch.object(ValidationError, "__lt__", spec=True, new=_lt)
def test_messages(name, schema, instance, expected, validator):
    validator = validator.evolve(schema=schema)
    errs = list(validator.iter_errors(instance))
    assert sorted(errs) == sorted(expected), f"{name!r} returned {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.functions.If import If


@pytest.fixture(scope="module")
def rule():
    rule = If()
    yield rule


@pytest.fixture
def template():
    return {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Conditions": {
            "IsUsEast1": {"Fn::Equals": [{"Ref": "AWS::Region"}, "us-east-1"]}
        },
        "Parameters": {
            "MyParameter": {
                "Type": "String",
                "Default": "foobar",
            },
        },
        "Resources": {},
    }


@pytest.fixture
def validator(cfn, context):
    validator = CfnTemplateValidator({}).extend(
        validators={
            "fn_if": If().fn_if,
        }
    )
    return validator(
        context=context,
        cfn=cfn,
        schema={},
    )


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::If",
            {"Fn::If": ["IsUsEast1", "foo", "bar"]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::If with to many arguments",
            {"Fn::If": ["IsUsEast1", "foo", "bar", "key"]},
            {"type": "string"},
            [
                ValidationError(
                    "expected maximum item count: 3, found: 4",
                    path=deque(["Fn::If"]),
                    schema_path=deque(["cfnContext", "schema", "maxItems"]),
                    validator="fn_if",
                ),
            ],
        ),
        (
            "Invalid Fn::If with bad first element",
            {"Fn::If": ["IsUsEast1", {"foo": "bar"}, "bar"]},
            {"type": "string"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string'",
                    path=deque(["Fn::If", 1]),
                    schema_path=deque(["type"]),
                    validator="type",
                ),
            ],
        ),
        (
            "Invalid Fn::If with bad condition",
            {"Fn::If": [{"Ref": "MyParameter"}, "foo", "bar"]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Ref': 'MyParameter'} is not one of ['IsUsEast1']",
                    path=deque(["Fn::If", 0]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            0,
                            "cfnContext",
                            "schema",
                            "dynamicValidation",
                            "enum",
                        ]
                    ),
                    validator="fn_if",
                ),
                ValidationError(
                    "{'Ref': 'MyParameter'} is not of type 'string'",
                    path=deque(["Fn::If", 0]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            0,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_if",
                ),
            ],
        ),
        (
            "Invalid Fn::If with bad second element",
            {"Fn::If": ["IsUsEast1", "foo", {"foo": "bar"}]},
            {"type": "string"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'string'",
                    path=deque(["Fn::If", 2]),
                    schema_path=deque(["type"]),
                    validator="type",
                ),
            ],
        ),
        (
            "Invalid Fn::If a condition that doesn't exist",
            {"Fn::If": ["foo", True, False]},
            {"type": "boolean"},
            [
                ValidationError(
                    "'foo' is not one of ['IsUsEast1']",
                    path=deque(["Fn::If", 0]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            0,
                            "cfnContext",
                            "schema",
                            "dynamicValidation",
                            "enum",
                        ]
                    ),
                    validator="fn_if",
                ),
            ],
        ),
        (
            "Invalid path",
            {"Fn::If": ["IsUsEast1", True, {"Fn::If": ["IsUsEast1", "foo", False]}]},
            {"type": "boolean"},
            [
                ValidationError(
                    (
                        "['Fn::If', 1] is not reachable. When setting condition "
                        "'IsUsEast1' to True from current status False"
                    ),
                    path=deque(["Fn::If", 2, "Fn::If", 1]),
                    schema_path=deque(["fn_if"]),
                    validator="fn_if",
                ),
                ValidationError(
                    ("'foo' is not of type 'boolean'"),
                    path=deque(["Fn::If", 2, "Fn::If", 1]),
                    schema_path=deque(["fn_if", "type"]),
                    validator="type",
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_if(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.RefFormat import RefFormat


@pytest.fixture(scope="module")
def rule():
    rule = RefFormat()
    yield rule


@pytest.fixture
def template():
    return {
        "Resources": {
            "MyBucket": {"Type": "AWS::S3::Bucket"},
            "MyVpc": {"Type": "AWS::EC2::VPC"},
            "MyCustomResource": {"Type": "Custom::MyResource"},
        },
    }


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Ref with a good format",
            "MyVpc",
            {"format": "AWS::EC2::VPC.Id"},
            [],
        ),
        (
            "Valid Ref with a good format",
            "MyCustomResource",
            {"format": "AWS::EC2::VPC.Id"},
            [],
        ),
        (
            "Invalid Ref with a bad format",
            "MyVpc",
            {"format": "AWS::EC2::Image.Id"},
            [
                ValidationError(
                    (
                        "{'Ref': 'MyVpc'} with formats "
                        "['AWS::EC2::VPC.Id'] does not match "
                        "destination format of 'AWS::EC2::Image.Id'"
                    ),
                    rule=RefFormat(),
                )
            ],
        ),
        (
            "Invalid Ref with a resource with no format",
            "MyBucket",
            {"format": "AWS::EC2::Image.Id"},
            [
                ValidationError(
                    (
                        "{'Ref': 'MyBucket'} does not match "
                        "destination format of 'AWS::EC2::Image.Id'"
                    ),
                    rule=RefFormat(),
                )
            ],
        ),
        (
            "Invalid Ref to non existent resource",
            "DNE",
            {"format": "AWS::EC2::Image.Id"},
            [],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.validate(validator, schema, instance, schema))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""


class BaseFormatter:
    """Base Formatter class"""

    def _format(self, match):
        """Format the specific match"""

    def print_matches(self, matches, rules, config):
        """Output all the matches"""
        if not matches:
            return None

        # Output each match on a separate line by default
        output = []
        for match in matches:
            output.append(self._format(match))

        return "\n".join(output)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class RestApiMixingDefinitions(CfnLintKeyword):
    id = "W3660"
    shortdesc = "Validate if multiple resources are modifying a Rest API definition"
    description = (
        "When using AWS::ApiGateway::RestApi with 'Body' or 'BodyS3Location' "
        "the resource handler will use PutRestApi with mode overwrite. "
        "Depending on how resources are updated the IaC template will "
        "drift and create orphaned resources."
    )
    tags = ["resources", "apigateway"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::ApiGateway::RestApi/Properties/Body",
                "Resources/AWS::ApiGateway::RestApi/Properties/BodyS3Location",
            ],
        )
        self._mix_types = [
            "AWS::ApiGateway::Method",
            "AWS::ApiGateway::Model",
            "AWS::ApiGateway::Resource",
            "AWS::ApiGateway::GatewayResponse",
            "AWS::ApiGateway::RequestValidator",
            "AWS::ApiGateway::Authorizer",
        ]

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        if validator.cfn.graph is None:  # pragma: no cover
            return  # pragma: no cover

        if not len(validator.context.path.path) > 3:
            return

        resource_name = validator.context.path.path[1]
        key = validator.context.path.path[3]

        unique_sources: set[str] = set()
        for source, _ in validator.cfn.graph.graph.in_edges(resource_name):
            if source not in validator.context.resources:
                continue
            if validator.context.resources[source].type in self._mix_types:
                if source not in unique_sources:
                    yield ValidationError(
                        message=(
                            f"Defining {key!r} with a relation to resource {source!r} "
                            f"of type {validator.context.resources[source].type!r} "
                            "may result in drift and orphaned resources"
                        ),
                        rule=self,
                    )
                    unique_sources.add(source)



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import Any

import cfnlint.data.schemas.extensions.aws_sqs_queue
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class QueueProperties(CfnLintJsonSchema):
    id = "E3501"
    shortdesc = "Validate SQS queue properties are valid"
    description = (
        "Depending on if the queue is FIFO or not the "
        "properties and allowed values change. "
        "This rule validates properties and values based on the queue type."
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-sqs-queue.html"
    tags = ["resources", "sqs"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::SQS::Queue/Properties",
            ],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_sqs_queue,
                filename="properties.json",
            ),
            all_matches=True,
        )

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        for err in super().validate(validator, keywords, instance, schema):
            if err.schema is False:
                err.message = (
                    f"Additional properties are not allowed ({err.path[-1]!r} "
                    "was unexpected)"
                )

            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from unittest import TestCase

from cfnlint.context import Context, Path
from cfnlint.helpers import FUNCTIONS
from cfnlint.jsonschema import CfnTemplateValidator, ValidationError
from cfnlint.rules.resources.iam.ResourcePolicy import ResourcePolicy


class TestResourcePolicy(TestCase):
    """Test IAM identity Policies"""

    def setUp(self):
        """Setup"""
        self.rule = ResourcePolicy()

    def test_object_basic(self):
        """Test Positive"""
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                )
            )
        )

        policy = {"Version": "2012-10-18"}

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 2, errs)
        self.assertEqual(
            errs[0].message, "'2012-10-18' is not one of ['2008-10-17', '2012-10-17']"
        )
        self.assertListEqual(list(errs[0].path), ["Version"])
        self.assertEqual(errs[1].message, "'Statement' is a required property")
        self.assertListEqual(list(errs[1].path), [])

    def test_object_multiple_effect(self):
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                )
            )
        )

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "NotAction": "*",
                    "Action": [
                        "cloudformation:*",
                    ],
                    "Resource": "*",
                    "Principal": {
                        "AWS": [
                            "arn:aws:iam::123456789012:root",
                            "999999999999",
                            "arn:aws:iam::cloudfront:user/CloudFront Origin Access Identity E1234ABCDE12AB",
                        ],
                        "CanonicalUser": "79a59df900b949e55d96a1e698fbacedfd6e09d98eacf8f8d5218e7cd47ef2be",
                    },
                    "Condition": {
                        "Null": {
                            "s3:x-amz-server-side-encryption": ["false"],
                            "aws:TagKeys": "false",
                        }
                    },
                }
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 2, errs)
        self.assertEqual(
            errs[0].message,
            ("Only one of ['Action', 'NotAction'] is a required property"),
        )
        self.assertEqual(
            errs[1].message,
            ("Only one of ['Action', 'NotAction'] is a required property"),
        )
        self.assertIn(
            ["Statement", 0, "NotAction"], [list(errs[0].path), list(errs[1].path)]
        )
        self.assertIn(
            ["Statement", 0, "Action"], [list(errs[0].path), list(errs[1].path)]
        )

    def test_object_statements(self):
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                functions=FUNCTIONS,
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                ),
            )
        )

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "NotAllow",
                    "Action": [
                        "cloudformation:Describe*",
                        "cloudformation:List*",
                        "cloudformation:Get*",
                    ],
                    "Resource": [
                        {
                            "Fn::Sub": "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                        },
                        {
                            "NotValid": [
                                "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                            ]
                        },
                    ],
                }
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 3, errs)
        self.assertEqual(
            errs[0].message,
            "Only one of ['Principal', 'NotPrincipal'] is a required property",
        )
        self.assertEqual(errs[1].message, "'NotAllow' is not one of ['Allow', 'Deny']")
        self.assertListEqual(list(errs[1].path), ["Statement", 0, "Effect"])
        self.assertEqual(
            errs[2].message,
            "{'NotValid': ['arn:${AWS::Partition}:iam::123456789012:role/object-role']} is not of type 'string'",
        )
        self.assertListEqual(list(errs[2].path), ["Statement", 0, "Resource", 1])

    def test_string_statements(self):
        """Test Positive"""
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                )
            )
        )

        # ruff: noqa: E501
        policy = """
            {
                "Version": "2012-10-18",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Action": [
                            "cloudformation:Describe*",
                            "cloudformation:List*",
                            "cloudformation:Get*"
                        ],
                        "Resource": [
                            "*",
                            {"Fn::Sub": ["arn:${AWS::Partition}:iam::123456789012/role/string-role"]}
                        ]
                    }
                ]
            }
        """

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertEqual(len(errs), 3, errs)
        self.assertEqual(
            errs[0].message,
            "Only one of ['Principal', 'NotPrincipal'] is a required property",
        )
        self.assertEqual(
            errs[1].message,
            "{'Fn::Sub': ['arn:${AWS::Partition}:iam::123456789012/role/string-role']} is not of type 'string'",
        )
        self.assertListEqual(list(errs[1].path), ["Statement", 0, "Resource", 1])
        self.assertEqual(
            errs[2].message, "'2012-10-18' is not one of ['2008-10-17', '2012-10-17']"
        )
        self.assertListEqual(list(errs[2].path), ["Version"])

    def test_principal_wildcard(self):
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                functions=FUNCTIONS,
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                ),
            )
        )

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": {
                        "Fn::Sub": "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                    },
                    "Principal": "*",
                },
                {
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": {
                        "Fn::Sub": "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                    },
                    "Principal": {
                        "AWS": "*",
                    },
                },
                {
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": {
                        "Fn::Sub": "arn:${AWS::Partition}:iam::123456789012:role/object-role"
                    },
                    "Principal": {"Fn::Sub": "*"},
                },
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertListEqual(errs, [])

    def test_assumed_role(self):
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                functions=FUNCTIONS,
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                ),
            )
        )

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "arn:aws:s3:::bucket",
                    "Principal": {
                        "AWS": "arn:aws:sts::123456789012:assumed-role/rolename/rolesessionname"
                    },
                },
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertListEqual(errs, [])

    def test_duplicate_sid(self):
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                functions=FUNCTIONS,
                path=Path(
                    cfn_path=deque(["Resources", "AWS::S3::BucketPolicy", "Properties"])
                ),
            )
        )

        policy = {
            "Version": "2012-10-17",
            "Statement": [
                {
                    "Sid": "A",
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "arn:aws:s3:::bucket",
                    "Principal": {
                        "AWS": "arn:aws:sts::123456789012:assumed-role/rolename/rolesessionname"
                    },
                },
                {
                    "Sid": "A",
                    "Effect": "Allow",
                    "Action": "*",
                    "Resource": "arn:aws:s3:::bucket",
                    "Principal": {
                        "AWS": "arn:aws:sts::123456789012:assumed-role/rolename/rolesessionname"
                    },
                },
            ],
        }

        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertListEqual(errs, [])

        # Fail on SNS topic
        validator = CfnTemplateValidator({}).evolve(
            context=Context(
                functions=FUNCTIONS,
                path=Path(
                    cfn_path=deque(["Resources", "AWS::SNS::TopicPolicy", "Properties"])
                ),
            )
        )
        errs = list(
            self.rule.validate(
                validator=validator, policy=policy, schema={}, policy_type=None
            )
        )
        self.assertListEqual(
            errs,
            [
                ValidationError(
                    "array items are not unique for keys ['Sid']",
                    validator="uniqueKeys",
                    schema_path=deque(["properties", "Statement", "uniqueKeys"]),
                    path=deque(["Statement"]),
                    rule=ResourcePolicy(),
                )
            ],
        )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.codepipeline.PipelineArtifactCounts import (
    PipelineArtifactCounts,
)


@pytest.fixture
def rule():
    rule = PipelineArtifactCounts()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            [],
            [],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "Source",
                    "Owner": "ThirdParty",
                    "Provider": "GitHub",
                    "Version": "1",
                },
                "OutputArtifacts": [{"Name": "MyApp"}],
            },
            [],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "FOO",
                    "Owner": "FOO",
                    "Provider": "FOO",
                    "Version": "1",
                },
            },
            [],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "FOO",
                    "Owner": "AWS",
                    "Provider": "FOO",
                    "Version": "1",
                },
            },
            [],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "Deploy",
                    "Owner": "AWS",
                    "Provider": "FOO",
                    "Version": "1",
                },
            },
            [],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "Source",
                    "Owner": "ThirdParty",
                    "Provider": "GitHub",
                    "Version": "1",
                },
                "OutputArtifacts": [{"Name": "Foo"}, {"Name": "Bar"}],
            },
            [
                ValidationError(
                    (
                        "expected maximum item count: 1, found: 2 "
                        "when using {'Owner': 'ThirdParty', "
                        "'Category': 'Source', 'Provider': 'GitHub'}"
                    ),
                    rule=PipelineArtifactCounts(),
                    validator="maxItems",
                    path=deque(["OutputArtifacts"]),
                    schema_path=deque(["properties", "OutputArtifacts", "maxItems"]),
                )
            ],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "Build",
                    "Owner": "AWS",
                    "Provider": "CodeBuild",
                    "Version": "1",
                },
            },
            [
                ValidationError(
                    (
                        "'InputArtifacts' is a required "
                        "property when using {'Owner': 'AWS', "
                        "'Category': 'Build', 'Provider': 'CodeBuild'}"
                    ),
                    rule=PipelineArtifactCounts(),
                    validator="required",
                    path=deque([]),
                    schema_path=deque(["required"]),
                )
            ],
        ),
        (
            {
                "Name": "Github",
                "ActionTypeId": {
                    "Category": "Deploy",
                    "Owner": "AWS",
                    "Provider": "S3",
                    "Version": "1",
                },
                "OutputArtifacts": [{"Name": "Foo"}, {"Name": "Bar"}],
                "InputArtifacts": [{"Name": "Foo"}, {"Name": "Bar"}],
            },
            [
                ValidationError(
                    (
                        "expected maximum item count: 1, found: 2 when "
                        "using {'Owner': 'AWS', 'Category': "
                        "'Deploy', 'Provider': 'S3'}"
                    ),
                    rule=PipelineArtifactCounts(),
                    validator="maxItems",
                    path=deque(["InputArtifacts"]),
                    schema_path=deque(["properties", "InputArtifacts", "maxItems"]),
                ),
                ValidationError(
                    (
                        "expected maximum item count: 0, found: 2 "
                        "when using {'Owner': 'AWS', 'Category': "
                        "'Deploy', 'Provider': 'S3'}"
                    ),
                    rule=PipelineArtifactCounts(),
                    validator="maxItems",
                    path=deque(["OutputArtifacts"]),
                    schema_path=deque(["properties", "OutputArtifacts", "maxItems"]),
                ),
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.resources.sqs.QueueProperties import QueueProperties


@pytest.fixture(scope="module")
def rule():
    rule = QueueProperties()
    yield rule


@pytest.mark.parametrize(
    "instance,expected",
    [
        (
            {
                "FifoQueue": True,
            },
            [],
        ),
        (
            {
                "FifoQueue": True,
                "QueueName": "test.fifo",
            },
            [],
        ),
        (
            {
                "FifoQueue": True,
                "QueueName": "test",
            },
            [
                ValidationError(
                    "'test' does not match '^.*\\\\.fifo$'",
                    rule=QueueProperties(),
                    path=deque(["QueueName"]),
                    validator="pattern",
                    schema_path=deque(
                        [
                            "allOf",
                            0,
                            "then",
                            "then",
                            "properties",
                            "QueueName",
                            "pattern",
                        ]
                    ),
                )
            ],
        ),
        (
            {
                "ContentBasedDeduplication": True,
            },
            [
                ValidationError(
                    (
                        "Additional properties are not allowed "
                        "('ContentBasedDeduplication' was unexpected)"
                    ),
                    rule=QueueProperties(),
                    path=deque(["ContentBasedDeduplication"]),
                    validator=None,
                    schema_path=deque(
                        ["allOf", 1, "then", "properties", "ContentBasedDeduplication"]
                    ),
                )
            ],
        ),
    ],
)
def test_validate(instance, expected, rule, validator):
    errs = list(rule.validate(validator, "", instance, {}))

    assert errs == expected, f"Expected {expected} got {errs}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.helpers import get_value_from_path
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class BucketTieringConfiguration(CfnLintKeyword):
    id = "E3061"
    shortdesc = "Validate the days for tierings in IntelligentTieringConfigurations"
    description = (
        "When using AWS::S3::Bucket to configure IntelligentTieringConfigurations "
        "the Tierings have minimum and maximum values"
    )
    source_url = "https://docs.aws.amazon.com/AmazonS3/latest/userguide/about-object-ownership.html"
    tags = ["resources", "s3"]

    def __init__(self):
        super().__init__(
            [
                "Resources/AWS::S3::Bucket/Properties/IntelligentTieringConfigurations/*/Tierings/*"
            ]
        )

        self._tierings = {
            "ARCHIVE_ACCESS": {
                "minimum": 90,
                "maximum": 730,
            },
            "DEEP_ARCHIVE_ACCESS": {
                "minimum": 180,
                "maximum": 730,
            },
        }

    def validate(
        self, validator: Validator, _, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:

        for archive_access, archive_access_validator in get_value_from_path(
            validator, instance, deque(["AccessTier"])
        ):
            if not validator.is_type(archive_access, "string"):
                continue
            for days, days_validator in get_value_from_path(
                archive_access_validator, instance, deque(["Days"])
            ):
                if validator.is_type(days, "string"):
                    try:
                        days = int(days)
                    except Exception:
                        return

                if not validator.is_type(days, "integer"):
                    continue

                days_validator = validator.evolve(
                    schema=self._tierings.get(archive_access, {})
                )

                for err in days_validator.iter_errors(days):
                    err.path = deque(["Days"])
                    err.rule = self
                    yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from typing import TYPE_CHECKING, Any, List, Protocol, Union

if TYPE_CHECKING:
    from cfnlint.rules import RuleMatch

RuleMatches = List["RuleMatch"]
Path = List[Union[str, int]]


class CheckValueFn(Protocol):
    def __call__(self, value: Any, path: Path, **kwargs: Any) -> List["RuleMatch"]: ...



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque
from typing import Any

from cfnlint.jsonschema import ValidationError, ValidationResult
from cfnlint.jsonschema.protocols import Validator
from cfnlint.rules import CloudFormationLintRule


class SubNotJoin(CloudFormationLintRule):
    id = "I1022"
    shortdesc = "Use Sub instead of Join"
    description = (
        "Prefer a sub instead of Join when using a join delimiter that is empty"
    )
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-sub.html"
    tags = ["functions", "sub", "join"]

    def _check_element(self, element):
        if isinstance(element, dict):
            if len(element) == 1:
                for key, value in element.items():
                    if key in ["Fn::Sub"]:
                        if not isinstance(value, str):
                            return False
                    elif key not in ["Ref", "Fn::GetAtt"]:
                        return False

        return True

    def _check_elements(self, elements):
        if not isinstance(elements, list):
            return False

        for element in elements:
            if not self._check_element(element):
                return False

        return True

    def validate(
        self, validator: Validator, s: Any, instance: Any, schema: Any
    ) -> ValidationResult:
        if validator.cfn.is_cdk_template():
            return

        key = list(instance.keys())[0]
        value = instance.get(key)

        if value[0] != "":
            return

        if self._check_elements(value[1]):
            yield ValidationError(
                ("Prefer using Fn::Sub over Fn::Join with an empty delimiter"),
                path=deque([key, 0]),
                rule=self,
            )



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

import cfnlint.data.schemas.extensions.aws_rds_dbcluster
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.rules.jsonschema.CfnLintJsonSchema import CfnLintJsonSchema, SchemaDetails


class DbClusterMultiAz(CfnLintJsonSchema):
    id = "E3692"
    shortdesc = "Validate Multi-AZ DB cluster configuration"
    description = (
        "When creating a Multi-AZ DB Cluster there are required fields and "
        "the allowed values are different"
    )
    tags = ["resources"]
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-rds-dbcluster.html#cfn-rds-dbcluster-engineversion"

    def __init__(self) -> None:
        super().__init__(
            keywords=["Resources/AWS::RDS::DBCluster/Properties"],
            schema_details=SchemaDetails(
                module=cfnlint.data.schemas.extensions.aws_rds_dbcluster,
                filename="multiaz.json",
            ),
            all_matches=True,
        )

    # DBClusterInstanceClass is required
    # AllocatedStorage is required
    # Iops is required
    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        for err in super().validate(validator, keywords, instance, schema):
            if err.schema is False:
                err.message = (
                    "Additional properties are not allowed "
                    f"{err.path[0]!r} when creating Multi-AZ cluster"
                )

            yield err



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.jsonschema._keywords_cfn import (
    _raw_type,
    _raw_type_strict,
    additionalProperties,
    cfn_type,
    cfnContext,
    dynamicValidation,
)


@pytest.mark.parametrize(
    "instance,schema,context_params,expected",
    [
        # Function allowed in context
        (
            {"Ref": "MyParam"},
            {"cfnContext": {"functions": ["Ref"], "schema": {"type": "object"}}},
            {},
            [],
        ),
        # Function not allowed in context
        (
            {"Fn::GetAtt": ["MyResource", "Arn"]},
            {"cfnContext": {"functions": ["Ref"], "schema": {"type": "string"}}},
            {},
            [
                ValidationError(
                    "{'Fn::GetAtt': ['MyResource', 'Arn']} is not of type 'string'",
                    path=deque([]),
                    schema_path=deque(["schema", "type"]),
                    validator="type",
                )
            ],
        ),
        # Valid schema in context
        (
            "test",
            {"cfnContext": {"functions": [], "schema": {"type": "string"}}},
            {},
            [],
        ),
        # Invalid schema in context
        (
            123,
            {"cfnContext": {"functions": [], "schema": {"type": "string"}}},
            {},
            [
                ValidationError(
                    "123 is not of type 'string'",
                    path=deque([]),
                    schema_path=deque(["schema", "type"]),
                    validator="type",
                )
            ],
        ),
        # Test with pseudoParameters
        (
            "AWS::Region",
            {
                "cfnContext": {
                    "functions": [],
                    "schema": {"type": "string"},
                    "pseudoParameters": ["AWS::Region", "AWS::AccountId"],
                }
            },
            {},
            [],
        ),
        # Test with references
        (
            {"Ref": "MyResource"},
            {
                "cfnContext": {
                    "functions": ["Ref"],
                    "schema": {"type": "object"},
                    "references": ["Parameters"],
                }
            },
            {},
            [],
        ),
        # Test with references including Resources
        (
            {"Ref": "MyResource"},
            {
                "cfnContext": {
                    "functions": ["Ref"],
                    "schema": {"type": "object"},
                    "references": ["Parameters", "Resources"],
                }
            },
            {},
            [],
        ),
        # Test with function object without $ref
        (
            {"Ref": "MyParam"},
            {
                "cfnContext": {
                    "functions": {"key": "value"},  # Object without $ref
                    "schema": {"type": "object"},
                }
            },
            {},
            [],
        ),
        # Test with null functions
        (
            {"Ref": "MyParam"},
            {"cfnContext": {"functions": None, "schema": {"type": "object"}}},
            {},
            [],
        ),
        # Test with null pseudoParameters
        (
            "AWS::Region",
            {
                "cfnContext": {
                    "functions": [],
                    "schema": {"type": "string"},
                    "pseudoParameters": None,
                }
            },
            {},
            [],
        ),
    ],
)
def test_cfn_context(instance, schema, context_params, expected, validator):
    if context_params:
        validator = validator.evolve(context=validator.context.evolve(**context_params))

    errs = list(cfnContext(validator, schema["cfnContext"], instance, schema))
    assert errs == expected


@pytest.mark.parametrize(
    "instance,schema,mock_resolver,expected",
    [
        # Test with functions as object with $ref
        (
            {"Ref": "MyParam"},
            {
                "cfnContext": {
                    "functions": {"$ref": "#/definitions/functions"},
                    "schema": {"type": "object"},
                }
            },
            {"#/definitions/functions": ["Ref"]},
            [],
        ),
    ],
)
def test_cfn_context_with_refs(instance, schema, mock_resolver, expected, validator):
    # Mock resolver for $ref resolution
    class MockResolver:
        def resolve(self, ref):
            return "", mock_resolver.get(ref, {})

    validator = validator.evolve(resolver=MockResolver())
    errs = list(cfnContext(validator, schema["cfnContext"], instance, schema))
    assert errs == expected


@pytest.mark.parametrize(
    "instance,schema,functions,expected",
    [
        # Test case where function matches and should be allowed
        (
            {"Ref": "MyParam", "Fn::GetAtt": ["Resource", "Attr"]},
            {"additionalProperties": False, "properties": {"Ref": {}}},
            ["Ref", "Fn::GetAtt"],
            [
                ValidationError(
                    (
                        "Additional properties are not allowed "
                        "('Fn::GetAtt' was unexpected)"
                    ),
                    path=deque(["Fn::GetAtt"]),
                )
            ],
        ),
        # Test case where function matches and should be allowed
        (
            {"Foo": True, "Fn::Transform": {}},
            {"additionalProperties": False, "properties": {"Foo": {}}},
            ["Fn::Transform"],
            [],
        ),
        # Test with empty instance
        (
            {},
            {"additionalProperties": False, "properties": {"Ref": {}}},
            ["Ref"],
            [],
        ),
        # Test with function in context.functions
        (
            {"Ref": "MyParam"},
            {"additionalProperties": False, "properties": {}},
            ["Ref"],
            [],
        ),
    ],
)
def test_additional_properties_with_functions(
    instance, schema, functions, expected, validator
):
    # Set up validator with functions
    validator = validator.evolve(context=validator.context.evolve(functions=functions))
    errs = list(additionalProperties(validator, False, instance, schema))
    assert errs == expected


@pytest.mark.parametrize(
    "template,path,instance,schema,context_params,expected",
    [
        # Transform check fails when transform is missing
        (
            {},
            {},
            "any_value",
            {"dynamicValidation": {"transformCheck": "AWS::LanguageExtensions"}},
            {},
            [
                ValidationError(
                    (
                        "Transform 'AWS::LanguageExtensions' is "
                        "required but not present in the template"
                    ),
                    path=deque([]),
                )
            ],
        ),
        # Transform check passes when transform is present
        (
            {"Transform": "AWS::LanguageExtensions"},
            {},
            "any_value",
            {"dynamicValidation": {"transformCheck": "AWS::LanguageExtensions"}},
            {},
            [],
        ),
        # Valid condition reference
        (
            {"Conditions": {"Condition1": {}, "Condition2": {}}},
            {},
            "Condition1",
            {"dynamicValidation": {"context": "conditions"}},
            {},
            [],
        ),
        # Invalid condition reference
        (
            {"Conditions": {"Condition1": {}, "Condition2": {}}},
            {},
            "InvalidCondition",
            {"dynamicValidation": {"context": "conditions"}},
            {},
            [
                ValidationError(
                    "'InvalidCondition' is not one of ['Condition1', 'Condition2']",
                    schema_path=deque(["enum"]),
                    path=deque([]),
                    validator="enum",
                )
            ],
        ),
        # Path check passes with matching pattern
        (
            {},
            {"path": ["Resources", "MyResource", "Properties", "InstanceType"]},
            "any_value",
            {
                "dynamicValidation": {
                    "pathCheck": "Resources/MyResource/Properties/InstanceType"
                }
            },
            {},
            [],
        ),
        # Path check fails with non-matching pattern
        (
            {},
            {"path": ["Parameters", "MyParam"]},
            "any_value",
            {"dynamicValidation": {"pathCheck": "Resources/MyResource/Properties"}},
            {},
            [
                ValidationError(
                    (
                        "'Parameters/MyParam' does not match "
                        "'^Resources/MyResource/Properties.*$'"
                    ),
                    schema_path=deque(["pattern"]),
                    path=deque([]),
                    validator="pattern",
                )
            ],
        ),
        # Test with non-object dynamicValidation value
        (
            {},
            {},
            "test",
            {"dynamicValidation": "not-an-object"},
            {},
            [],
        ),
        # Test with mappings context
        (
            {},
            {},
            "MyMapping",
            {"dynamicValidation": {"context": "mappings"}},
            {
                "mappings": type(
                    "obj",
                    (object,),
                    {"maps": {}, "mappings": {}},  # Empty mappings for this test
                )()
            },
            [
                ValidationError(
                    "'MyMapping' is not one of []",
                    schema_path=deque(["enum"]),
                    path=deque([]),
                    validator="enum",
                )
            ],
        ),
        # Test with context that's not a string
        (
            {},
            {},
            "test",
            {"dynamicValidation": {"context": 123}},  # Not a string
            {},
            [],
        ),
        # Test with context that doesn't match any collection
        (
            {},
            {},
            "test",
            {"dynamicValidation": {"context": "unknown_context"}},
            {},
            [],
        ),
    ],
    indirect=["template", "path"],
)
def test_dynamic_validation(
    template, path, instance, schema, context_params, expected, validator
):
    if context_params:
        validator = validator.evolve(context=validator.context.evolve(**context_params))

    errs = list(
        dynamicValidation(validator, schema["dynamicValidation"], instance, schema)
    )
    assert errs == expected


@pytest.mark.parametrize(
    "type_name,instance,strict,expected_result",
    [
        # _raw_type tests (strict=False)
        ("object", {}, False, True),
        ("array", [], False, True),
        ("null", None, False, True),
        ("string", "test", False, True),
        ("string", {}, False, False),
        ("string", [], False, False),
        ("number", 123, False, True),
        ("number", "123", False, True),
        ("number", True, False, False),
        ("integer", 123, False, True),
        ("integer", "123", False, True),
        ("integer", True, False, False),
        ("boolean", True, False, True),
        ("boolean", "true", False, True),
        ("boolean", "false", False, True),
        # _raw_type_strict tests (strict=True)
        ("object", {}, True, True),
        ("string", "test", True, True),
        ("number", 123, True, True),
        ("number", "123", True, False),
        ("integer", 123, True, True),
        ("integer", "123", True, False),
        ("boolean", True, True, True),
        ("boolean", "true", True, False),
        # Edge cases for number type
        ("number", "not-a-number", False, False),  # Will cause ValueError in float()
        ("number", {"key": "value"}, False, False),  # Will cause TypeError in float()
        # Edge cases for integer type
        ("integer", "not-an-integer", False, False),  # Will cause ValueError in int()
        ("integer", {"key": "value"}, False, False),  # Will cause TypeError in int()
        # Edge case for boolean type
        ("boolean", "not-a-boolean", False, False),  # Not in BOOLEAN_STRINGS
        # Additional edge cases
        ("unknown-type", "test", False, False),  # Unknown type
        ("null", "null", False, False),  # String "null" is not None
    ],
)
def test_raw_type_functions(validator, type_name, instance, strict, expected_result):
    if strict:
        assert _raw_type_strict(validator, type_name, instance) == expected_result
    else:
        assert _raw_type(validator, type_name, instance) == expected_result


@pytest.mark.parametrize(
    "instance,type_value,strict_types,expected",
    [
        # Test with strict_types=True
        (
            "123",
            "number",
            True,
            [
                ValidationError(
                    "'123' is not of type 'number'",
                    path=deque([]),
                )
            ],
        ),
        # Test with strict_types=False
        (
            "123",
            "number",
            False,
            [],
        ),
        # Test with multiple types
        (
            "test",
            ["string", "number"],
            True,
            [],
        ),
        # Test with invalid type
        (
            123,
            "string",
            True,
            [
                ValidationError(
                    "123 is not of type 'string'",
                    path=deque([]),
                )
            ],
        ),
        # Test with complex type combinations
        (
            {"key": "value"},
            ["string", "number", "boolean", "object"],
            True,
            [],
        ),
        # Test with non-matching complex types
        (
            {"key": "value"},
            ["string", "number", "boolean"],
            True,
            [
                ValidationError(
                    "{'key': 'value'} is not of type 'string', 'number', 'boolean'",
                    path=deque([]),
                )
            ],
        ),
        # Test with null instance
        (
            None,
            "null",
            True,
            [],
        ),
        # Test with boolean instance and string type
        (
            True,
            "string",
            False,
            [],  # With strict_types=False, True can be considered a string
        ),
    ],
)
def test_cfn_type_function(instance, type_value, strict_types, expected, validator):
    validator = validator.evolve(
        context=validator.context.evolve(strict_types=strict_types)
    )
    errs = list(cfn_type(validator, type_value, instance, {}))
    assert errs == expected



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.functions.Join import Join


@pytest.fixture(scope="module")
def rule():
    rule = Join()
    yield rule


@pytest.mark.parametrize(
    "name,instance,schema,expected",
    [
        (
            "Valid Fn::Join with array",
            {"Fn::Join": ["foo", ["bar"]]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Join with wrong output type",
            {"Fn::Join": ["foo", ["bar"]]},
            {"type": "array"},
            [
                ValidationError(
                    "{'Fn::Join': ['foo', ['bar']]} is not of type 'array'",
                    path=deque([]),
                    schema_path=deque([]),
                    validator="fn_join",
                ),
            ],
        ),
        (
            "Invalid Fn::Join is NOT a array",
            {"Fn::Join": "foo"},
            {"type": "string"},
            [
                ValidationError(
                    "'foo' is not of type 'array'",
                    path=deque(["Fn::Join"]),
                    schema_path=deque(["cfnContext", "schema", "type"]),
                    validator="fn_join",
                ),
            ],
        ),
        (
            "Invalid Fn::Join using a function for delimiter",
            {"Fn::Join": [{"Ref": "MyResource"}, ["bar"]]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Ref': 'MyResource'} is not of type 'string'",
                    path=deque(["Fn::Join", 0]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            0,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_join",
                ),
            ],
        ),
        (
            "Invalid Fn::Join using an invalid function",
            {"Fn::Join": ["-", {"foo": "bar"}]},
            {"type": "string"},
            [
                ValidationError(
                    "{'foo': 'bar'} is not of type 'array'",
                    path=deque(["Fn::Join", 1]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            1,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_join",
                ),
            ],
        ),
        (
            "Invalid Fn::Join using an invalid CFN function",
            {"Fn::Join": ["-", {"Fn::Join": ["-", "bar"]}]},
            {"type": "string"},
            [
                ValidationError(
                    "{'Fn::Join': ['-', 'bar']} is not of type 'array'",
                    path=deque(["Fn::Join", 1]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            1,
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_join",
                ),
            ],
        ),
        (
            "Valid Fn::Join with a valid function",
            {"Fn::Join": ["-", {"Fn::Split": ["-", "bar"]}]},
            {"type": "string"},
            [],
        ),
        (
            "Valid Fn::Join with a valid item function",
            {"Fn::Join": ["-", [{"Fn::Sub": "bar"}]]},
            {"type": "string"},
            [],
        ),
        (
            "Invalid Fn::Join with an invalid function",
            {"Fn::Join": ["-", [{"Fn::Split": ["-", {"Ref": "MyResource"}]}]]},
            {"type": "string"},
            [
                ValidationError(
                    (
                        "{'Fn::Split': ['-', {'Ref': 'MyResource'}]} "
                        "is not of type 'string'"
                    ),
                    path=deque(["Fn::Join", 1, 0]),
                    schema_path=deque(
                        [
                            "cfnContext",
                            "schema",
                            "prefixItems",
                            1,
                            "cfnContext",
                            "schema",
                            "items",
                            "cfnContext",
                            "schema",
                            "type",
                        ]
                    ),
                    validator="fn_join",
                ),
            ],
        ),
    ],
)
def test_validate(name, instance, schema, expected, rule, validator):
    errs = list(rule.fn_join(validator, schema, instance, {}))
    assert errs == expected, f"Test {name!r} got {errs!r}"



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

from typing import Any

from cfnlint.context.conditions import Unsatisfiable
from cfnlint.jsonschema import ValidationResult, Validator
from cfnlint.jsonschema.exceptions import ValidationError
from cfnlint.rules.jsonschema.CfnLintKeyword import CfnLintKeyword


class PipelineArtifactNames(CfnLintKeyword):
    id = "E3701"
    shortdesc = "Validate input and output artifact names are used properly"
    description = (
        "When using AWS::CodePipeline::Pipeline InputArtifacts names "
        "have to be previously used OutputArtifact names. Additionally, "
        "the OutputArtifacts names have to be unique"
    )
    tags = ["resources", "codepipeline"]

    def __init__(self) -> None:
        super().__init__(
            keywords=[
                "Resources/AWS::CodePipeline::Pipeline/Properties/Stages/*/Actions/*/InputArtifacts/*/Name",
                "Resources/AWS::CodePipeline::Pipeline/Properties/Stages/*/Actions/*/OutputArtifacts/*/Name",
            ],
        )
        self._output_artifact_names: dict[str, list[tuple[str, dict]]] = {}

    def initialize(self, cfn):
        self._output_artifact_names = {}
        return super().initialize(cfn)

    def validate(
        self, validator: Validator, keywords: Any, instance: Any, schema: dict[str, Any]
    ) -> ValidationResult:
        if not validator.is_type(instance, "string"):
            return

        resource_name = validator.context.path.path[1]
        if not isinstance(resource_name, str):
            return

        if resource_name not in self._output_artifact_names:
            self._output_artifact_names[resource_name] = []

        if "OutputArtifacts" in validator.context.path.path:
            for output_name, output_condition in self._output_artifact_names[
                resource_name
            ]:
                if output_name != instance:
                    continue
                try:
                    validator.evolve(
                        context=validator.context.evolve(
                            conditions=validator.context.conditions.evolve(
                                output_condition
                            )
                        )
                    )
                    yield ValidationError(
                        f"{instance!r} is already a defined 'OutputArtifact' Name",
                        rule=self,
                    )
                    return
                except Unsatisfiable:
                    pass

            self._output_artifact_names[resource_name].append(
                (instance, validator.context.conditions.status)
            )
        elif "InputArtifacts" in validator.context.path.path:
            for output_name, output_condition in self._output_artifact_names[
                resource_name
            ]:
                if output_name != instance:
                    continue
                try:
                    validator.evolve(
                        context=validator.context.evolve(
                            conditions=validator.context.conditions.evolve(
                                output_condition
                            )
                        )
                    )
                    break
                except Unsatisfiable:
                    pass
            else:
                yield ValidationError(
                    f"{instance!r} is not previously defined as an 'OutputArtifact'",
                    rule=self,
                )
                return



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from __future__ import annotations

import cfnlint.data.schemas.other.functions
from cfnlint.rules.functions._BaseFn import BaseFn, SchemaDetails


class Condition(BaseFn):
    """Check And Condition Function Logic"""

    id = "E8007"
    shortdesc = "Check Condition structure for validity"
    description = "Check Condition has a value of another condition"
    source_url = "https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference-conditions.html#intrinsic-function-reference-conditions-and"
    tags = ["functions", "condition"]

    def __init__(self) -> None:
        super().__init__(
            "Condition",
            ("boolean",),
            schema_details=SchemaDetails(
                cfnlint.data.schemas.other.functions, "condition.json"
            ),
        )
        self.condition = self.validate



"""
Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
SPDX-License-Identifier: MIT-0
"""

from collections import deque

import pytest

from cfnlint.jsonschema import ValidationError
from cfnlint.rules.mappings.Configuration import Configuration


@pytest.fixture(scope="module")
def rule():
    rule = Configuration()
    yield rule


@pytest.mark.parametrize(
    "name,instance,expected",
    [
        (
            "Empty is okay",
            {"MyMapping": {"Key": {"Value": "Foo"}}},
            [],
        ),
        (
            "Good key name",
            {"MyMapping": {"Foo-Bar": {"Value": "Foo"}}},
            [],
        ),
        (
            "Wrong type",
            [],
            [
                ValidationError(
                    "[] is not of type 'object'",
                    validator="type",
                    schema_path=deque(["type"]),
                    rule=Configuration(),
                )
            ],
        ),
        (
            "Bad mapping type",
            {"MyMapping": []},
            [
                ValidationError(
                    "[] is not of type 'object'",
                    validator="type",
                    schema_path=deque(["patternProperties", "^[a-zA-Z0-9]+$", "type"]),
                    rule=Configuration(),
                    path=deque(["MyMapping"]),
                )
            ],
        ),
        (
            "Bad key type",
            {"MyMapping": {"Key": []}},
            [
                ValidationError(
                    "[] is not of type 'object'",
                    validator="type",
                    schema_path=deque(
                        [
                            "patternProperties",
                            "^[a-zA-Z0-9]+$",
                            "patternProperties",
                            "^[a-zA-Z0-9.-]+$",
                            "type",
                        ]
                    ),
                    rule=Configuration(),
                    path=deque(
                        [
                            "MyMapping",
                            "Key",
                        ]
                    ),
                )
            ],
        ),
        (
            "Bad value type",
            {"MyMapping": {"Key": {"Value": {}}}},
            [
                ValidationError(
                    (
                        "{} is not of type 'string', 'boolean', "
                        "'integer', 'number', 'array'"
                    ),
                    validator="type",
                    schema_path=deque(
                        [
                            "patternProperties",
                            "^[a-zA-Z0-9]+$",
                            "patternProperties",
                            "^[a-zA-Z0-9.-]+$",
                            "patternProperties",
                            "^[a-zA-Z0-9]+$",
                            "type",
                        ]
                    ),
                    rule=Configuration(),
                    path=deque(["MyMapping", "Key", "Value"]),
                )
            ],
        ),
        (
            "Bad Name",
            {"MyMapping-1": {"Key": {"Value": "Foo"}}},
            [
                ValidationError(
                    "'MyMapping-1' does not match any of the regexes: '^[a-zA-Z0-9]+$'",
                    validator="additionalProperties",
                    schema_path=deque(["additionalProperties"]),
                    rule=Configuration(),
                    path=deque(["MyMapping-1"]),
                )
            ],
        ),
        (
            "Long key name",
            {"a" * 256: {"Key": {"Value": "Foo"}}},
            [
                ValidationError(
                    "expected maximum length: 255, found: 256",
                    validator="maxLength",
                    schema_path=deque(["propertyNames", "maxLength"]),
                    rule=None,  # empty because we didn't load child rules
                    path=deque(["a" * 256]),
                )
            ],
        ),
    ],
)
def test_validate(name, instance, expected, rule, validator):
    errs = list(rule.validate(validator, {}, instance, {}))

    assert errs == expected, f"Test {name!r} got {errs!r}"
